#pylint: disable=too-many-lines
# -*- coding: utf-8 -*-

"""Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import Tuple, Dict, Any, List, Set, Generator, Optional, TYPE_CHECKING, Union

# we disable those because of our automatic conversion to classmethod/staticmethod
# mypy: disable-error-code="misc, override, call-arg, arg-type"

from tile import Tile, TileList, elements, Meld, MeldList
from tilesource import TileSource
from common import IntDict
from wind import East
from message import Message
from query import Query
from permutations import Permutations

if TYPE_CHECKING:
    from hand import Hand
    from game import PlayingGame
    from tile import Tiles
    from intelligence import AIDefaultAI, DiscardCandidates


# pylint:disable=missing-function-docstring,missing-class-docstring
# the class and method names are mostly self explaining, we do not
# need docstringss
# pylint: disable=no-self-argument,no-member
# pylint: disable=too-many-function-args, unused-argument, arguments-differ


class RuleCode:

    """Parent for all RuleCode classes. A RuleCode class can be used to
    define the behaviour of a Rule. Classes Rule and RuleCode
    are separate because
    - different rulesets may have a Rule with the same name
      but with different behaviour
    - different rulesets may use different names for the same rule
    - the RuleCode class should be as short and as concise
      as possible because this is the important part about
      implementing a new ruleset, and it is the most error prone.

    All methods in RuleCode classes will automatically be converted
    into staticmethods or classmethods if the 1st arg is named 'cls'.

    winningTileCandidates(cls, hand:'Hand'):
        All rules for going MahJongg must have such a method.
        This is used to find all winning hands which only need
        one tile: The calling hands (after calling)

    """

    cache : Tuple[str, ...] = ()

    options:Dict[str, Any] = {}

    def appliesToHand(self, hand:'Hand') ->bool:
        """returns true if this applies to hand"""
        return False

class MJRule(RuleCode):

    def computeLastMelds(hand:'Hand') ->MeldList:
        """return all possible last melds"""
        return MeldList()

    def shouldTry(hand:'Hand', maxMissing:int=10) ->bool:
        return True

    def rearrange(cls, hand:'Hand', rest:Union[TileList,
        List[Tile]]) ->Generator[Tuple[MeldList, TileList], None, None]:
        """rest is a list of those tiles that can still
        be rearranged: No declared melds and no bonus tiles.
        done is already arranged, do not change this.
        TODO: also return how many tiles are missing for winning"""
        permutations = Permutations(rest)
        for variantMelds in permutations.variants:
            yield variantMelds, TileList()


class DragonPungKong(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isPungKong and meld.isDragonMeld


class ExposedMinorPung(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isPung and meld[0].isMinor and meld.isExposed


class ExposedTerminalsPung(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isExposed and meld[0].isTerminal and meld.isPung


class ExposedHonorsPung(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isExposed and meld.isHonorMeld and meld.isPung


class ExposedMinorKong(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isExposed and meld[0].isMinor and meld.isKong


class ExposedTerminalsKong(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isExposed and meld[0].isTerminal and meld.isKong


class ExposedHonorsKong(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isExposed and meld.isHonorMeld and meld.isKong


class ConcealedMinorPung(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isConcealed and meld[0].isMinor and meld.isPung


class ConcealedTerminalsPung(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isConcealed and meld[0].isTerminal and meld.isPung


class ConcealedHonorsPung(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isConcealed and meld.isHonorMeld and meld.isPung


class ConcealedMinorKong(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isConcealed and meld[0].isMinor and meld.isKong


class ConcealedTerminalsKong(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isConcealed and meld[0].isTerminal and meld.isKong


class ConcealedHonorsKong(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isConcealed and meld.isHonorMeld and meld.isKong


class OwnWindPungKong(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        assert hand
        return meld[0].value is hand.ownWind

    def mayApplyToMeld(meld:Meld) ->bool:
        """for meld rules which depend on context like hand.ownWind, we want
        to know if there could be a context where this rule applies. See
        Meld.rules.
        NOTE: If a rulecode class has mayApplyToMeld, its appliesToMeld can
        assume that mayApplyToMeld has already been checked."""
        return meld.isPungKong and meld.isWindMeld


class OwnWindPair(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        assert hand
        return meld[0].value is hand.ownWind

    def mayApplyToMeld(meld:Meld) ->bool:
        return meld.isPair and meld.isWindMeld


class RoundWindPungKong(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        assert hand
        return meld[0].value is hand.roundWind

    def mayApplyToMeld(meld:Meld) ->bool:
        return meld.isPungKong and meld.isWindMeld


class RoundWindPair(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        assert hand
        return meld[0].value is hand.roundWind

    def mayApplyToMeld(meld:Meld) ->bool:
        return meld.isPair and meld.isWindMeld


class DragonPair(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isDragonMeld and meld.isPair


class LastTileCompletesPairMinor(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        if hand.lastMeld is None:
            return False
        return hand.lastMeld.isPair and hand.lastTile.isMinor


class Flower(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isSingle and meld.group == Tile.flower


class Season(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        return meld.isSingle and meld.group == Tile.season


class LastTileCompletesPairMajor(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        if hand.lastMeld is None:
            return False
        return bool(hand.lastMeld) and hand.lastMeld.isPair and hand.lastTile.isMajor


class LastFromWall(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return bool(hand.lastTile) and hand.lastTile.isConcealed


class ZeroPointHand(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return not any(x.meld for x in hand.usedRules if x.meld and len(x.meld) > 1)


class NoChow(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return not any(x.isChow for x in hand.melds)


class OnlyConcealedMelds(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return not any((x.isExposed and not x.isClaimedKong) for x in hand.melds)


class FalseColorGame(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        dwSet = set(Tile.honors)
        return bool(dwSet & hand.suits) and len(hand.suits - dwSet) == 1


class TrueColorGame(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return len(hand.suits) == 1 and hand.suits < set(Tile.colors)


class Purity(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return (len(hand.suits) == 1 and hand.suits < set(Tile.colors)
                and not any(x.isChow for x in hand.melds))


class ConcealedTrueColorGame(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        if len(hand.suits) != 1 or hand.suits >= set(Tile.colors):
            return False
        return not any((x.isExposed and not x.isClaimedKong) for x in hand.melds)


class OnlyMajors(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return all(x.isMajor for x in hand.tiles)


class OnlyHonors(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return all(x.isHonor for x in hand.tiles)


class HiddenTreasure(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return (not any(((x.isExposed and not x.isClaimedKong) or x.isChow) for x in hand.melds)
                and bool(hand.lastTile) and hand.lastTile.isConcealed
                and len(hand.melds) == 5)


class BuriedTreasure(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return (len(hand.suits - set(Tile.honors)) == 1
                and sum(x.isPung for x in hand.melds) == 4
                and all((x.isPung and x.isConcealed) or x.isPair for x in hand.melds))


class AllTerminals(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return all(x.isTerminal for x in hand.tiles)


class StandardMahJongg(MJRule):

    cache : Tuple[str, ...] = ('appliesToHand',)

    def computeLastMelds(hand:'Hand') ->MeldList:
        """return all possible last melds"""
        return MeldList(x for x in hand.melds if hand.lastTile in x and len(x) < 4)

    def appliesToHand(hand:'Hand') ->bool:
        """winner rules are not yet applied to hand"""
        # pylint: disable=too-many-return-statements
        # too many return statements
        if len(hand.melds) != 5:
            return False
        if any(len(x) not in (2, 3, 4) for x in hand.melds):
            return False
        if any(x.isRest or x.isKnitted for x in hand.melds):
            return False
        if sum(x.isChow for x in hand.melds) > hand.ruleset.maxChows:
            return False
        if hand.arranged is None:
            # this is only Hand.__arrange
            return True
        assert hand.score
        if hand.score.total() < hand.ruleset.minMJPoints:
            return False
        if hand.score.doubles >= hand.ruleset.minMJDoubles:
            # shortcut
            return True
        # but maybe we have enough doubles by winning:
        doublingWinnerRules = sum(
            x.rule.score.doubles for x in hand.matchingWinnerRules())
        return hand.score.doubles + doublingWinnerRules >= hand.ruleset.minMJDoubles

    def fillChow(group:str, values:List[int]) ->Set[Tile]:
        val0, val1 = values
        if val0 + 1 == val1:
            if val0 == 1:
                return {Tile(group, val0 + 2)}
            if val0 == 8:
                return {Tile(group, val0 - 1)}
            return {Tile(group, val0 - 1), Tile(group, val0 + 2)}
        assert val0 + 2 == val1, f'group:{group} values:{values}'
        return {Tile(group, val0 + 1)}

    def winningTileCandidates(cls, hand:'Hand') ->Set[Tile]:
        # pylint: disable=too-many-locals,too-many-return-statements,too-many-branches,too-many-statements
        if len(hand.melds) > 7:
            # hope 7 is sufficient, 6 was not
            return set()
        if not hand.tilesInHand:
            return set()
        inHand = [x.exposed for x in hand.tilesInHand]
        result = inHand[:]
        pairs = 0
        isolated = 0
        maxChows = hand.ruleset.maxChows - \
            sum(x.isChow for x in hand.declaredMelds)
        # TODO: does not differentiate between maxChows == 1 and maxChows > 1
        # test with kajonggtest and a ruleset where maxChows == 2
        if maxChows < 0:
            return set()
        if maxChows == 0:
            check_tiles = set(inHand)
        else:
            check_tiles = set(inHand) & elements.honors
        for tileName in check_tiles:
            count = inHand.count(tileName)
            if count == 1:
                isolated += 1
            elif count == 2:
                pairs += 1
            else:
                for _ in range(count):
                    result.remove(tileName)
        if maxChows:
            if pairs > 2 or isolated > 2 or (pairs > 1 and isolated > 1):
                # this is not a calling hand
                return set()
        else:
            if pairs + isolated > 2:
                return set()
        if maxChows == 0:
            return set(result)
        melds = []
        for group in sorted(hand.suits & set(Tile.colors)):
            values = sorted(x.value for x in result if x.group == group)
            changed = True
            while (changed and len(values) > 2
                   and values.count(values[0]) == 1
                   and values.count(values[1]) == 1
                   and values.count(values[2]) == 1):
                changed = False
                if values[0] + 2 == values[2] and (len(values) == 3 or values[3] > values[0] + 3):
                    # logDebug('removing first 3 from %s' % values)
                    meld = Tile(group, values[0]).chow
                    assert meld
                    # must be a pylint bug. meld is TileList is list
                    for pair in meld:
                        result.remove(pair)
                    melds.append(meld)
                    values = values[3:]
                    changed = True
                elif values[0] + 1 == values[1] and values[2] > values[0] + 2:
                    # logDebug('found incomplete chow at start of %s' %
                    # values)
                    return cls.fillChow(group, values[:2])
            changed = True
            while (changed and len(values) > 2
                   and values.count(values[-1]) == 1
                   and values.count(values[-2]) == 1
                   and values.count(values[-3]) == 1):
                changed = False
                if values[-1] - 2 == values[-3] and (len(values) == 3 or values[-4] < values[-1] - 3):
                    meld = Tile(group, values[-3]).chow
                    assert meld
                    # must be a pylint bug. meld is TileList is list
                    for pair in meld:
                        result.remove(pair)
                    melds.append(meld)
                    values = values[:-3]
                    changed = True
                elif values[-1] - 1 == values[-2] and values[-3] < values[-1] - 2:
                    # logDebug('found incomplete chow at end of %s' % values)
                    return cls.fillChow(group, values[-2:])

            if len(values) % 3 == 0:
                # adding a 4th, 7th or 10th tile with this color can not let us win,
                # so we can exclude this color from the candidates
                result = [x for x in result if x.group != group]
                continue
            valueSet = set(values)
            if len(values) == 4 and len(values) == len(valueSet):
                if values[0] + 3 == values[-1]:
                    # logDebug('seq4 in %s' % hand.tilesInHand)
                    return {Tile(group, values[0]), Tile(group, values[-1])}
            if len(values) == 7 and len(values) == len(valueSet):
                if values[0] + 6 == values[6]:
                    # logDebug('seq7 in %s' % hand.tilesInHand)
                    return {Tile(group, values[x]) for x in (0, 3, 6)}
            if len(values) == 1:
                # only a pair of this value is possible
                # logDebug('need pair')
                return {Tile(group, values[0]).concealed}
            if len(valueSet) == 1:
                # no chow reachable, only pair/pung
                continue
            singles = {x for x in valueSet
                       if values.count(x) == 1
                       and not {x - 1, x - 2, x + 1, x + 2} & valueSet}
            isolated += len(singles)
            if isolated > 1:
                # this is not a calling hand
                return set()
            if len(values) == 2 and len(valueSet) == 2:
                # exactly two adjacent values: must be completed to Chow
                if maxChows == 0:
                    # not a calling hand
                    return set()
                # logDebug('return fillChow for %s' % values)
                return cls.fillChow(group, values)
            if (len(values) == 4 and len(valueSet) == 2
                    and values[0] == values[1] and values[2] == values[3]):
                return {Tile(group, values[0]), Tile(group, values[2])}
            if maxChows:
                for value in valueSet:
                    if value > 1:
                        result.append(Tile(group, value - 1))
                    if value < 9:
                        result.append(Tile(group, value + 1))
        return set(result)


class StandardMahJongg16(StandardMahJongg):

    def appliesToHand(hand:'Hand') ->bool:
        """winner rules are not yet applied to hand"""
        # pylint: disable=too-many-return-statements
        # too many return statements
        if len(hand.melds) != 6:
            return False
        if any(len(x) not in (2, 3, 4) for x in hand.melds):
            return False
        if any(x.isRest or x.isKnitted for x in hand.melds):
            return False
        if sum(x.isChow for x in hand.melds) > hand.ruleset.maxChows:
            return False
        if hand.arranged is None:
            # this is only Hand.__arrange
            return True
        assert hand.score
        if hand.score.total() < hand.ruleset.minMJPoints:
            return False
        if hand.score.doubles >= hand.ruleset.minMJDoubles:
            # shortcut
            return True
        # but maybe we have enough doubles by winning:
        doublingWinnerRules = sum(
            x.rule.score.doubles for x in hand.matchingWinnerRules())
        return hand.score.doubles + doublingWinnerRules >= hand.ruleset.minMJDoubles

class SquirmingSnake(StandardMahJongg):
    cache : Tuple[str, ...] = ()

    def appliesToHand(hand:'Hand') ->bool:
        cacheKey = (hand.ruleset.standardMJRule.__class__, 'appliesToHand')
        std = hand.ruleCache.get(cacheKey, None)
        if std is False:
            return False
        if len(hand.suits) != 1 or hand.suits >= set(Tile.colors):
            return False
        values = hand.values
        if values.count(1) < 3 or values.count(9) < 3:
            return False
        pairs = [x for x in (2, 5, 8) if values.count(x) == 2]
        if len(pairs) != 1:
            return False
        return len(set(values)) == len(values) - 5

    def winningTileCandidates(hand:'Hand') ->Set[Tile]:
        """they have already been found by the StandardMahJongg rule"""
        return set()


class WrigglingSnake(MJRule):

    def shouldTry(hand:'Hand', maxMissing:int=3) ->bool:
        if hand.declaredMelds:
            return False
        return (len({x.exposed for x in hand.tiles}) + maxMissing > 12
                and all(not x.isChow for x in hand.declaredMelds))

    def computeLastMelds(hand:'Hand') ->MeldList:
        if not hand.lastTile:
            return MeldList()
        _ = Tile(hand.lastTile)
        return MeldList(_.pair) if hand.lastTile.value == 1 else MeldList(_.single)

    def winningTileCandidates(hand:'Hand') ->Set[Tile]:
        suits = hand.suits.copy()
        if Tile.wind not in suits or Tile.dragon in suits or len(suits) > 2:
            return set()
        suits -= {Tile.wind}
        if not suits:
            return set()
        group = suits.pop()
        values = set(hand.values)
        if len(values) < 12:
            return set()
        if len(values) == 12:
            # one of 2..9 or a wind is missing
            if hand.values.count(1) < 2:
                # and the pair of 1 is incomplete too
                return set()
            return (elements.winds | {Tile(group, x) for x in range(2, 10)}) \
                - {x.exposed for x in hand.tiles}
        # pair of 1 is not complete
        return {Tile(group, '1')}

    def rearrange(cls, hand:'Hand', rest:Union[TileList,
        List[Tile]]) ->Generator[Tuple[MeldList, TileList], None, None]:
        melds = MeldList()
        for tileName in rest[:]:
            if rest.count(tileName) >= 2:
                melds.append(tileName.pair)
                rest.remove(tileName)
                rest.remove(tileName)
            elif rest.count(tileName) == 1:
                melds.append(tileName.single)
                rest.remove(tileName)
        yield melds, rest

    def appliesToHand(hand:'Hand') ->bool:
        if hand.declaredMelds:
            return False
        suits = hand.suits.copy()
        if Tile.wind not in suits:
            return False
        suits -= {Tile.wind}
        if len(suits) != 1 or suits >= set(Tile.colors):
            return False
        if hand.values.count(1) != 2:
            return False
        return len(set(hand.values)) == 13


class CallingHand(RuleCode):

    def appliesToHand(cls, hand:'Hand') ->bool:  # pylint:disable=arguments-renamed
        for callHand in hand.callingHands:
            used = (x.rule.__class__ for x in callHand.usedRules)
            assert hasattr(cls, 'limitHand')
            if cls.limitHand in used:
                return True
        return False


class TripleKnitting(MJRule):

    def computeLastMelds(cls, hand:'Hand') ->MeldList:
        """return all possible last melds"""
        if not hand.lastTile:
            return MeldList()
        triples, rest = cls.findTriples(hand)
        assert len(rest) == 2
        triples.append(Meld(rest))
        return MeldList(Meld(x) for x in triples if hand.lastTile in x)

    def claimness(cls, hand:'Hand', discard:Optional[Tile]) ->IntDict:
        result = IntDict()
        if cls.shouldTry(hand):
            result[Message.Pung] = -999
            result[Message.Kong] = -999
            result[Message.Chow] = -999
        return result

    def weigh(cls, aiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
        if cls.shouldTry(candidates.hand):
            _, rest = cls.findTriples(candidates.hand)
            for candidate in candidates:
                if candidate.group in Tile.honors:
                    candidate.keep -= 50
                if rest.count(candidate.tile) > 1:
                    candidate.keep -= 10
        return candidates

    def rearrange(cls, hand:'Hand', rest:Union[TileList,
        List[Tile]]) ->Generator[Tuple[MeldList, TileList], None, None]:
        melds = MeldList()
        for triple in cls.findTriples(hand)[0]:
            melds.append(triple)
            rest.remove(triple[0])
            rest.remove(triple[1])
            rest.remove(triple[2])
        while len(rest) >= 2:
            for tile in sorted(set(rest)):
                value = tile.value
                suits = {x.group for x in rest if x.value == value}
                if len(suits) < 2:
                    yield melds, rest
                    return
                pair = (Tile(suits.pop(), value), Tile(suits.pop(), value))
                melds.append(Meld(sorted(pair)))
                rest.remove(pair[0])
                rest.remove(pair[1])
        yield melds, rest

    def appliesToHand(cls, hand:'Hand') ->bool:  # pylint:disable=arguments-renamed
        if any(x.isHonor for x in hand.tiles):
            return False
        if len(hand.declaredMelds) > 1:
            return False
        if hand.lastTile and hand.lastTile.isConcealed and hand.declaredMelds:
            return False
        triples, rest = cls.findTriples(hand)
        return (len(triples) == 4 and len(rest) == 2
                and rest[0].group != rest[1].group and rest[0].value == rest[1].value)

    def winningTileCandidates(cls, hand:'Hand') ->Set[Tile]:
        if hand.declaredMelds:
            return set()
        if any(x.isHonor for x in hand.tiles):
            return set()
        _, rest = cls.findTriples(hand)
        if len(rest) not in (1, 4):
            return set()
        result = list(
            Tile(x, y.value).concealed for x in Tile.colors for y in rest)
        for restTile in rest:
            result.remove(restTile)
        return set(result)

    def shouldTry(cls, hand:'Hand', maxMissing:int=3) ->bool:  # pylint:disable=arguments-renamed
        if hand.declaredMelds:
            return False
        tripleWanted = 4 - maxMissing // 3  # count triples
        tripleCount = len(cls.findTriples(hand)[0])
        return tripleCount >= tripleWanted

    def findTriples(hand:'Hand') ->Tuple[MeldList, TileList]:
        """return a list of triple knitted melds, including the mj triple.
        Also returns the remaining untripled tiles"""
        if hand.declaredMelds:
            if len(hand.declaredMelds) > 1:
                return MeldList(), TileList()
        result = MeldList()
        tilesS = [x.concealed for x in hand.tiles if x.lowerGroup == Tile.stone]
        tilesB = [x.concealed for x in hand.tiles if x.lowerGroup == Tile.bamboo]
        tilesC = [x.concealed for x in hand.tiles if x.lowerGroup == Tile.character]
        for tileS in tilesS[:]:
            tileB = Tile(Tile.bamboo, tileS.value).concealed
            tileC = Tile(Tile.character, tileS.value).concealed
            if tileB in tilesB and tileC in tilesC:
                tilesS.remove(tileS)
                tilesB.remove(tileB)
                tilesC.remove(tileC)
                result.append(tileS.knitted3)
        return result, TileList(tilesS + tilesB + tilesC)


class Knitting(MJRule):

    def computeLastMelds(cls, hand:'Hand') ->MeldList:
        """return all possible last melds"""
        if not hand.lastTile:
            return MeldList()
        couples, rest = cls.findCouples(hand)
        assert not rest, f'{hand.string}: couples={couples} rest={rest}'
        return MeldList(Meld(x) for x in couples if hand.lastTile in x)

    def claimness(cls, hand:'Hand', discard:Optional[Tile]) ->IntDict:
        result = IntDict()
        if cls.shouldTry(hand):
            result[Message.Pung] = -999
            result[Message.Kong] = -999
            result[Message.Chow] = -999
        return result

    def weigh(cls, aiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
        if cls.shouldTry(candidates.hand):
            for candidate in candidates:
                if candidate.group in Tile.honors:
                    candidate.keep -= 50
        return candidates

    def shouldTry(cls, hand:'Hand', maxMissing:int=4) ->bool:  # pylint:disable=arguments-renamed
        if hand.declaredMelds:
            return False
        pairWanted = 7 - maxMissing // 2  # count pairs
        pairCount = len(cls.findCouples(hand)[0])
        return pairCount >= pairWanted

    def appliesToHand(cls, hand:'Hand') ->bool:  # pylint:disable=arguments-renamed
        if any(x.isHonor for x in hand.tiles):
            return False
        if len(hand.declaredMelds) > 1:
            return False
        if hand.lastTile and hand.lastTile.isConcealed and hand.declaredMelds:
            return False
        return len(cls.findCouples(hand)[0]) == 7

    def winningTileCandidates(cls, hand:'Hand') ->Set[Tile]:
        if hand.declaredMelds:
            return set()
        if any(x.isHonor for x in hand.tiles):
            return set()
        couples, singleTile = cls.findCouples(hand)
        if len(couples) != 6:
            return set()
        if not singleTile:
            # single tile has wrong suit
            return set()
        assert len(singleTile) == 1
        tile = singleTile[0]
        otherSuit = (hand.suits - {tile.lowerGroup}).pop()
        otherTile = Tile(otherSuit, tile.value).concealed
        return {otherTile}

    def rearrange(cls, hand:'Hand', rest:Union[TileList,
        List[Tile]]) ->Generator[Tuple[MeldList, TileList], None, None]:
        melds = MeldList()
        for couple in cls.findCouples(hand, rest)[0]:
            if couple[0].isExposed:
                # this is the mj pair, lower after claiming
                continue
            melds.append(Meld(couple))
            rest.remove(couple[0])
            rest.remove(couple[1])
        yield melds, rest

    def findCouples(cls, hand:'Hand', pairs:TileList=TileList()) ->Tuple[MeldList, TileList]:
        """return a list of tuples, including the mj couple.
        Also returns the remaining uncoupled tiles IF they
        are of the wanted suits"""
        if hand.declaredMelds:
            if len(hand.declaredMelds) > 1 or len(hand.declaredMelds[0]) > 2:
                return MeldList(), TileList()
        result = MeldList()
        if not pairs:
            pairs = hand.tiles
        suits = cls.pairSuits(hand)
        if not suits:
            return MeldList(), TileList()
        tiles0 = TileList(x for x in pairs if x.lowerGroup == suits[0])
        tiles1 = TileList(x for x in pairs if x.lowerGroup == suits[1])
        for tile0 in tiles0[:]:
            if tile0.isExposed:
                tile1 = Tile(suits[1], tile0.value)
            else:
                tile1 = Tile(suits[1], tile0.value).concealed
            if tile1 in tiles1:
                tiles0.remove(tile0)
                tiles1.remove(tile1)
                result.append((tile0, tile1))
        return result, tiles0 + tiles1

    def pairSuits(hand:'Hand') ->str:
        """return a lowercase string with two suit characters. If no prevalence, returns ''"""
        suitCounts = [len([x for x in hand.tiles if x.lowerGroup == y]) for y in Tile.colors]
        minSuit = min(suitCounts)
        result = ''.join(x for idx, x in enumerate(Tile.colors) if suitCounts[idx] > minSuit)
        if len(result) == 2:
            return result
        return ''


class AllPairHonors(MJRule):

    def computeLastMelds(hand:'Hand') ->MeldList:
        return MeldList(Tile(hand.lastTile).pair)

    def claimness(cls, hand:'Hand', discard:Optional[Tile]) ->IntDict:
        result = IntDict()
        if cls.shouldTry(hand):
            result[Message.Pung] = -999
            result[Message.Kong] = -999
            result[Message.Chow] = -999
        return result

    def maybeCallingOrWon(hand:'Hand') ->bool:
        if any(x.value in Tile.minors for x in hand.tiles):
            return False
        return len(hand.declaredMelds) < 2

    def appliesToHand(cls, hand:'Hand') ->bool:  # pylint:disable=arguments-renamed
        if not cls.maybeCallingOrWon(hand):
            return False
        if len(set(hand.tiles)) != 7:
            return False
        return {len([x for x in hand.tiles if x == y]) for y in hand.tiles} == {2}

    def winningTileCandidates(cls, hand:'Hand') ->Set[Tile]:
        if not cls.maybeCallingOrWon(hand):
            return set()
        single = [x for x in hand.tiles if hand.tiles.count(x) == 1]
        if len(single) != 1:
            return set()
        return set(single)

    def shouldTry(hand:'Hand', maxMissing:int=4) ->bool:
        if hand.declaredMelds:
            return False
        tiles = [x.exposed for x in hand.tiles]
        pairCount = kongCount = 0
        for tile in elements.majors:
            count = tiles.count(tile)
            if count == 2:
                pairCount += 1
            elif count == 4:
                kongCount += 1
        pairWanted = 7 - maxMissing // 2  # count pairs
        result = pairCount >= pairWanted or (
            pairCount + kongCount * 2) > pairWanted
        return result

    def rearrange(cls, hand:'Hand', rest:Union[TileList,
        List[Tile]]) ->Generator[Tuple[MeldList, TileList], None, None]:
        melds = MeldList()
        for pair in sorted(set(rest) & elements.mAJORS):
            while rest.count(pair) >= 2:
                melds.append(pair.pair)
                rest.remove(pair)
                rest.remove(pair)
        yield melds, rest

    def weigh(aiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
        hand = candidates.hand
        if not AllPairHonors.shouldTry(hand):
            return candidates
        keep = 10
        for candidate in candidates:
            if candidate.value in Tile.minors:
                candidate.keep -= keep
            else:
                if candidate.occurrence == 3:
                    candidate.keep -= keep / 2
                else:
                    candidate.keep += keep
        return candidates


class FourfoldPlenty(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return len(hand.tiles) == 18


class ThreeGreatScholars(RuleCode):

    def appliesToHand(cls, hand:'Hand') ->bool:  # pylint:disable=arguments-renamed
        return (BigThreeDragons.appliesToHand(hand)
                and ('nochow' not in cls.options or not any(x.isChow for x in hand.melds)))


class BigThreeDragons(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return len([x for x in hand.melds if x.isDragonMeld and x.isPungKong]) == 3


class BigFourJoys(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return len([x for x in hand.melds if x.isWindMeld and x.isPungKong]) == 4


class LittleFourJoys(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        lengths = sorted(min(len(x), 3) for x in hand.melds if x.isWindMeld)
        return lengths == [2, 3, 3, 3]


class LittleThreeDragons(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return sorted(min(len(x), 3) for x in hand.melds if x.isDragonMeld) == [2, 3, 3]


class FourBlessingsHoveringOverTheDoor(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return len([x for x in hand.melds if x.isPungKong and x.isWindMeld]) == 4


class AllGreen(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return {x.exposed for x in hand.tiles} < elements.greenHandTiles


class LastTileFromWall(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return hand.lastSource is TileSource.LivingWall


class LastTileFromDeadWall(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return hand.lastSource is TileSource.DeadWall

    def selectable(hand:'Hand') ->bool:
        """for scoring game"""
        return hand.lastSource is TileSource.LivingWall


class IsLastTileFromWall(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return hand.lastSource is TileSource.LivingWallEnd

    def selectable(hand:'Hand') ->bool:
        """for scoring game"""
        return hand.lastSource is TileSource.LivingWall


class IsLastTileFromWallDiscarded(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return hand.lastSource is TileSource.LivingWallEndDiscard

    def selectable(hand:'Hand') ->bool:
        """for scoring game"""
        return hand.lastSource is TileSource.LivingWallDiscard


class RobbingKong(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return hand.lastSource is TileSource.RobbedKong

    def selectable(hand:'Hand') ->bool:
        """for scoring game"""
        return (hand.lastSource in (TileSource.RobbedKong, TileSource.LivingWall, TileSource.LivingWallDiscard)
                and bool(hand.lastTile )and hand.lastTile.group.islower()
                and [x.exposed for x in hand.tiles].count(hand.lastTile.exposed) < 2)


class GatheringPlumBlossomFromRoof(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return LastTileFromDeadWall.appliesToHand(hand) and hand.lastTile is Tile(Tile.stone, '5').concealed


class PluckingMoon(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return IsLastTileFromWall.appliesToHand(hand) and hand.lastTile is Tile(Tile.stone, '1').concealed


class ScratchingPole(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return RobbingKong.appliesToHand(hand) and hand.lastTile is Tile(Tile.bamboo, '2')


class StandardRotation(RuleCode):

    def rotate(game:'PlayingGame') ->bool:
        if game is None:
            return False
        if game.winner is None:
            return False
        return game.winner.wind is not East


class EastWonNineTimesInARow(RuleCode):
    nineTimes = 9

    def appliesToHand(cls, hand:'Hand') ->bool:  # pylint:disable=arguments-renamed
        return cls.appliesToGame(hand.player.game)

    def appliesToGame(game:'PlayingGame', needWins:Optional[int]=None) ->bool:
        if needWins is None:
            needWins = EastWonNineTimesInARow.nineTimes
            if game.isScoringGame():
                # we are only proposing for the last needed Win
                needWins -= 1
        if game.winner and game.winner.wind is East and game.point.notRotated >= needWins:
            eastMJCount = int(Query("select count(1) from score "
                                    f"where game={game.gameid} and won=1 and wind='E' "
                                    f"and player={int(game.players[East].nameid)} "
                                    f"and prevailing='{game.point.prevailing}'").records[0][0])
            return eastMJCount == needWins
        return False

    def rotate(cls, game:'PlayingGame') ->bool:
        return cls.appliesToGame(game, EastWonNineTimesInARow.nineTimes)


class GatesOfHeaven(StandardMahJongg):
    """as used for Classical Chinese BMJA.

    I believe that when they say a run of 2..8, they must
    all be concealed
    """

    cache = ()

# TODO: in BMJA, 111 and 999 must be concealed, we do not check this
    def computeLastMelds(hand:'Hand') ->MeldList:
        return MeldList(x for x in hand.melds if hand.lastTile in x)

    def shouldTry(hand:'Hand', maxMissing:Optional[int]=None) ->bool:
        if hand.declaredMelds:
            return False
        for suit in Tile.colors:
            count19 = sum(x.isTerminal for x in hand.tiles)
            suitCount = len([x for x in hand.tiles if x.lowerGroup == suit])
            if suitCount > 10 and count19 > 4:
                return True
        return False

    def appliesToHand(hand:'Hand') ->bool:
        if len(hand.suits) != 1 or hand.suits >= set(Tile.colors):
            return False
        if any(len(x) > 2 for x in hand.declaredMelds):
            return False
        values = hand.values
        if len(set(values)) < 9 or values.count(1) != 3 or values.count(9) != 3:
            return False
        values_list = list(values[3:-3])
        for value in Tile.minors:
            if value in values_list:
                values_list.remove(value)
        if len(values_list) != 1:
            return False
        surplus = values_list[0]
        return 1 < surplus < 9

    def winningTileCandidates(hand:'Hand') ->Set[Tile]:
        if hand.declaredMelds:
            return set()
        if len(hand.suits) != 1 or hand.suits >= set(Tile.colors):
            return set()
        values = hand.values
        if len(set(values)) < 9:
            return set()
        # we have something of all values
        if values.count(1) != 3 or values.count(9) != 3:
# TODO: we may get them from the wall but not by claim. Differentiate!
            return set()
        for suit in hand.suits:
            return {Tile(suit, x) for x in Tile.minors}
        return set()

    def rearrange(cls, hand:'Hand', rest:Union[TileList,
        List[Tile]]) ->Generator[Tuple[MeldList, TileList], None, None]:
        melds = MeldList()
        for suit in hand.suits & set(Tile.colors):
            for value in Tile.numbers:
                tile = Tile(suit, value).concealed
                if rest.count(tile) == 3 and tile.isTerminal:
                    melds.append(tile.pung)
                    rest.remove(tile)
                    rest.remove(tile)
                    rest.remove(tile)
                elif rest.count(tile) == 2:
                    melds.append(tile.pair)
                    rest.remove(tile)
                    rest.remove(tile)
                elif rest.count(tile) == 1:
                    melds.append(tile.single)
                    rest.remove(tile)
            break
        yield melds, rest


class NineGates(GatesOfHeaven):
    """as used for Classical Chinese DMJL"""

    def appliesToHand(hand:'Hand') ->bool:
        """last tile may also be 1 or 9"""
        if hand.declaredMelds:
            return False
        if len(hand.suits) != 1 or hand.suits >= set(Tile.colors):
            return False
        values = hand.values
        if len(set(values)) < 9:
            return False
        if values.count(1) != 3 or values.count(9) != 3:
            return False
        values_list = list(values[3:-3])
        for value in Tile.minors:
            if value in values_list:
                values_list.remove(value)
        if len(values_list) != 1:
            return False
        surplus = values_list[0]
        return bool(hand.lastTile) and surplus == hand.lastTile.value

    def winningTileCandidates(hand:'Hand') ->Set[Tile]:
        if hand.declaredMelds:
            return set()
        if len(hand.suits) != 1 or hand.suits >= set(Tile.colors):
            return set()
        values = hand.values
        if len(set(values)) < 9:
            return set()
        # we have something of all values
        if values.count(1) != 3 or values.count(9) != 3:
            return set()
        for suit in hand.suits:
            return {Tile(suit, x) for x in Tile.numbers}
        return set()


class ThirteenOrphans(MJRule):

    def computeLastMelds(hand:'Hand') ->MeldList:
        meldSize = hand.tilesInHand.count(hand.lastTile)
        if meldSize == 0:
            # the last tile is called and not yet in the hand
            return MeldList()
        return MeldList(Tile(hand.lastTile).meld(meldSize))

    def rearrange(cls, hand:'Hand', rest:Union[TileList,
        List[Tile]]) ->Generator[Tuple[MeldList, TileList], None, None]:
        melds = MeldList()
        for tileName in rest:
            if rest.count(tileName) >= 2:
                melds.append(tileName.pair)
                rest.remove(tileName)
                rest.remove(tileName)
            elif rest.count(tileName) == 1:
                melds.append(tileName.single)
                rest.remove(tileName)
        yield melds, rest

    def claimness(cls, hand:'Hand', discard:Optional[Tile]) ->IntDict:
        result = IntDict()
        if cls.shouldTry(hand):
            doublesCount = hand.doublesEstimate(discard)
# TODO: compute scoring for resulting hand. If it is high anyway,
# prefer pung over trying 13 orphans
            if doublesCount < 2 or cls.shouldTry(hand, 1):
                result[Message.Pung] = -999
                result[Message.Kong] = -999
                result[Message.Chow] = -999
        return result

    def appliesToHand(hand:'Hand') ->bool:
        return {x.exposed for x in hand.tiles} == elements.majors

    def winningTileCandidates(cls, hand:'Hand') ->Set[Tile]:
        if any(x in hand.values for x in Tile.minors):
            # no minors allowed
            return set()
        if not cls.shouldTry(hand, 1):
            return set()
        handTiles = {x.exposed for x in hand.tiles}
        missing = elements.majors - handTiles
        if not missing:
            # if all 13 tiles are there, we need any one of them:
            return elements.majors
        assert len(missing) == 1
        return missing

    def shouldTry(hand:'Hand', maxMissing:int=4) ->bool:
        # TODO: look at how many tiles there still are on the wall
        if hand.declaredMelds:
            return False
        handTiles = {x.exposed for x in hand.tiles}
        missing = elements.majors - handTiles
        if len(missing) > maxMissing:
            return False
        for missingTile in missing:
            if not hand.player.tileAvailable(missingTile, hand):
                return False
        return True

    def weigh(cls, aiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
        hand = candidates.hand
        assert hand
        if not cls.shouldTry(hand):
            return candidates
        handTiles = {x.exposed for x in hand.tiles}
        missing = elements.majors - handTiles
        havePair = False
        keep = (6 - len(missing)) * 5
        for candidate in candidates:
            if candidate.value in Tile.minors:
                candidate.keep -= keep
            else:
                if havePair and candidate.occurrence >= 2:
                    candidate.keep -= keep
                else:
                    candidate.keep += keep
                havePair = candidate.occurrence == 2
        return candidates


class OwnFlower(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        assert hand
        return meld[0].value is hand.ownWind

    def mayApplyToMeld(meld:Meld) ->bool:
        # pylint: disable=unsubscriptable-object
        # must be a pylint bug. meld is TileList is list
        return meld.isBonus and meld[0].group == Tile.flower


class OwnSeason(RuleCode):

    def appliesToMeld(hand:Optional['Hand'], meld:Meld) ->bool:
        assert hand
        return meld[0].value is hand.ownWind

    def mayApplyToMeld(meld:Meld) ->bool:
        # pylint: disable=unsubscriptable-object
        # must be a pylint bug. meld is TileList is list
        return meld.isBonus and meld[0].group == Tile.season


class OwnFlowerOwnSeason(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return sum(x.isBonus and x[0].value is hand.ownWind for x in hand.bonusMelds) == 2


class AllFlowers(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return len([x for x in hand.bonusMelds if x.group == Tile.flower]) == 4


class AllSeasons(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return len([x for x in hand.bonusMelds if x.group == Tile.season]) == 4


class ThreeConcealedPongs(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return len([x for x in hand.melds if (
            x.isConcealed or x.isClaimedKong) and x.isPungKong]) >= 3


class MahJonggWithOriginalCall(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return ('a' in hand.announcements
                and sum(x.isExposed for x in hand.melds) < 3)

    def selectable(hand:'Hand') ->bool:
        """for scoring game"""
        # one tile may be claimed before declaring OC and one for going MJ
        # the previous regex was too strict
        return sum(x.isExposed for x in hand.melds) < 3

    def claimness(hand:'Hand', discard:Optional[Tile]) ->IntDict:
        result = IntDict()
        player = hand.player
        if player.originalCall and player.mayWin:
            if player.originalCallingHand and player.originalCallingHand.chancesToWin():
                # winning with OriginalCall is still possible
                result[Message.Pung] = -999
                result[Message.Kong] = -999
                result[Message.Chow] = -999
            else:
                player.mayWin = False  # bad luck
        return result


class TwofoldFortune(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return 't' in hand.announcements

    def selectable(hand:'Hand') ->bool:
        """for scoring game"""
        kungs = [x for x in hand.melds if len(x) == 4]
        return len(kungs) >= 2


class BlessingOfHeaven(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        if hand.lastSource is not TileSource.East14th:
            return False
        if hand.ownWind is not East:
            return False
        if any(x.isExposed for x in hand.melds):
            return False
        assert hand.lastTile is Tile.none, f'{hand}: Blessing of Heaven: There can be no last tile'
        return True

    def selectable(hand:'Hand') ->bool:
        """for scoring game"""
        return (hand.ownWind is East
                and hand.lastSource in (TileSource.LivingWall, TileSource.LivingWallDiscard)
                and not hand.announcements - {'a'})


class BlessingOfEarth(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        if hand.lastSource is not TileSource.East14th:
            return False
        if hand.ownWind is East:
            return False
        return True

    def selectable(hand:'Hand') ->bool:
        """for scoring game"""
        return (hand.ownWind is not East
                and hand.lastSource in (TileSource.LivingWall, TileSource.LivingWallDiscard)
                and not hand.announcements - {'a'})


class LongHand(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return hand.lenOffset > 0 if not hand.won else hand.lenOffset > 1


class FalseDiscardForMJ(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return not hand.won

    def selectable(hand:'Hand') ->bool:
        """for scoring game"""
        return not hand.won


class DangerousGame(RuleCode):

    def appliesToHand(hand:'Hand') ->bool:
        return not hand.won

    def selectable(hand:'Hand') ->bool:
        """for scoring game"""
        return not hand.won


class LastOnlyPossible(RuleCode):

    """check if the last tile was the only one possible for winning"""
    def appliesToHand(hand:'Hand') ->bool:
        if not hand.lastTile:
            return False
        if any(hand.lastTile in x for x in hand.melds if len(x) == 4):
            # the last tile completed a Kong
            return False
        shortHand = hand - hand.lastTile
        return len(shortHand.callingHands) == 1



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2014 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import Dict, Union, Optional, TYPE_CHECKING, List, Tuple, cast, Iterator, overload

from qt import Qt, QPointF, QPoint, QRectF, QMimeData, QSize
from qt import QGraphicsRectItem, QSizePolicy, QFrame, QFont
from qt import QGraphicsView, QLabel, QObject
from qt import QColor, QPainter, QDrag, QPixmap, QStyleOptionGraphicsItem, QPen, QBrush
from qt import QFontMetrics, QGraphicsObject
from qt import QMenu, QCursor
from qt import QGraphicsSvgItem
from tileset import Tileset
from tile import Tile, elements, Meld, MeldList
from uitile import UITile, UIMeld
from guiutil import Painter, rotateCenter, sceneRotation
from animation import AnimationSpeed, animate, AnimatedMixin
from message import Message

from util import stack, uniqueList
from log import logDebug, logException
from mi18n import i18n, i18nc
from common import LIGHTSOURCES, Internal, Debug, isAlive, ReprMixin
from common import DrawOnTopMixin
from wind import Wind, East, NoWind

if TYPE_CHECKING:
    from qt import QGraphicsSceneDragDropEvent, QWidget, QSizeF
    from qt import QKeyEvent, QMouseEvent, QWheelEvent, QResizeEvent
    from uiwall import UIWallSide, UIWall
    from game import PlayingGame, Game
    from message import ClientMessage
    from scene import SceneWithFocusRect, ScoringScene
    from handboard import HandBoard


class WindDisc(DrawOnTopMixin, AnimatedMixin, QGraphicsObject, ReprMixin):  # type:ignore[misc]

    """a round wind tile"""

    roundWindColor = QColor(235, 235, 173)
    whiteColor = QColor('white')

    def __init__(self, wind:Wind, parent:Optional[QObject]=None) ->None:
        """generate new wind disc"""
        super().__init__()
        assert not parent
        assert isinstance(wind, Wind), f'wind {wind}  must be a real Wind but is {type(wind)}'
        self.__wind:Wind = wind
        self.__brush = self.whiteColor
        self.board:'UIWallSide'

    def debug_name(self) ->str:
        """for identification in animations"""
        return self.__wind.tile.name2()

    def moveDict(self) ->Dict[str, Union[QPointF, str, int]]:
        """a dict with attributes for the new position,
        normally pos, rotation and scale"""
        sideCenter = self.board.center()
        boardPos = QPointF(
            sideCenter.x() * 1.63,
            sideCenter.y() - self.boundingRect().height() / 2.0)
        scenePos = self.board.mapToScene(boardPos)
        return {'pos': scenePos, 'rotation': sceneRotation(self.board)}

    @property
    def wind(self) ->Wind:
        """our wind"""
        return self.__wind

    @property
    def is_prevailing(self) ->bool:
        """is this the prevailing wind?"""
        return self.__brush is self.roundWindColor

    @is_prevailing.setter
    def is_prevailing(self, value:bool) ->None:
        if isinstance(value, bool):
            newPrevailing = value
        else:
            newPrevailing = self.wind == Wind.all4[value % 4]
        self.__brush = self.roundWindColor if newPrevailing else self.whiteColor

    def paint(self, painter:Optional[QPainter], unusedOption:Optional[QStyleOptionGraphicsItem],
        unusedWidget:Optional['QWidget']=None) ->None:
        """paint the disc"""
        if painter:
            with Painter(painter):
                painter.setBrush(self.__brush)
                assert Internal.scene
                size = int(Internal.scene.windTileset.faceSize.height())
                ellRect = QRectF(QPointF(), QPointF(size, size))
                painter.drawEllipse(ellRect)
                renderer = Internal.scene.windTileset.renderer
                painter.translate(12, 12)
                painter.scale(0.60, 0.60)
                renderer.render(painter, self.wind.discSvgName, self.boundingRect())

    def boundingRect(self) ->QRectF:
        """define the part of the tile we want to see"""
        assert Internal.scene
        size = int(Internal.scene.windTileset.faceSize.height() * 1.1)
        return QRectF(QPointF(), QPointF(size, size))

    def __str__(self) ->str:
        """for debugging"""
        return f'WindDisc({self.debug_name()} x/y= {self.x():.1f}/{self.y():1f})'


class WindLabel(QLabel):

    """QLabel holding the wind tile"""

    def __init__(self, wind:Optional[Wind]=None, parent:Optional['QWidget']=None):
        super().__init__(parent)
        if wind is None:
            wind = East
        self.__wind:Wind
        self.__prevailing:Wind = NoWind
        self.wind = wind

    @property
    def wind(self) ->Wind:
        """the current wind on this label"""
        return self.__wind

    @wind.setter
    def wind(self, wind:Wind) ->None:
        """setting the wind also changes the pixmap"""
        if not hasattr(self, '__wind') or self.__wind != wind:
            self.__wind = wind
            self._refresh()

    @property
    def prevailing(self) ->Wind:
        """setting prevailing also changes graphics if needed"""
        return self.__prevailing

    @prevailing.setter
    def prevailing(self, value:Wind) ->None:
        """setting prevailing also changes graphics if needed"""
        if self.__prevailing != value:
            self.__prevailing = value
            self._refresh()

    def _refresh(self) ->None:
        """update graphics"""
        self.setPixmap(self.genWindPixmap())

    def genWindPixmap(self) ->QPixmap:
        """prepare wind tiles"""
        pwind = WindDisc(self.__wind)
        pwind.is_prevailing = self.__wind == self.__prevailing
        pMap = QPixmap(40, 40)
        pMap.fill(Qt.GlobalColor.transparent)
        painter = QPainter(pMap)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        painter.scale(0.40, 0.40)
        pwind.paint(painter, QStyleOptionGraphicsItem())
        for child in pwind.childItems():
            if isinstance(child, QGraphicsSvgItem):
                with Painter(painter):
                    painter.translate(child.mapToParent(0.0, 0.0))
                    child.paint(painter, QStyleOptionGraphicsItem())
        return pMap


class Board(QGraphicsRectItem, ReprMixin):

    """ a board with any number of positioned tiles"""
    # pylint: disable=too-many-instance-attributes

    penColor = QColor('black')

    arrows:List[Qt.Key] = [Qt.Key.Key_Left, Qt.Key.Key_Down, Qt.Key.Key_Up, Qt.Key.Key_Right]

    showShadowsBetweenRows = False

    def __init__(self, width:float, height:float, tileset:Tileset, boardRotation:int=0) ->None:
        super().__init__()
        self.uiTiles:List[UITile] = []
        self.isHandBoard:bool = False
        self._focusTile:Optional[UITile] = None
        self.__prevPos = 0.0
        self._noPen()
        self.tileDragEnabled = False
        self.setRotation(boardRotation)
        self._lightSource = 'NW'
        self.__xWidth:int = 0
        self.__xHeight:float = 0
        self.__yWidth:int = 0
        self.__yHeight:float = 0
        self.__fixedWidth:float = width
        self.__fixedHeight:float = height
        self._tileset = Tileset()
        self.tileset = tileset
        self.level:int = 0
        assert Internal.Preferences
        Internal.Preferences.addWatch('showShadows', self.showShadowsChanged)

    @property
    def empty(self) -> bool:
        """True if we have no uiTiles"""
        return not bool(self.uiTiles)

    def debug_name(self) ->str:
        """default board name, used for debugging messages"""
        return 'board'

    def __str__(self) ->str:
        """for debugging"""
        return self.debug_name()

    def setVisible(self, value:bool) ->None:
        """also update focusRect if it belongs to this board"""
        if self.scene() and isAlive(self):
            super().setVisible(value)
            cast('SceneWithFocusRect', self.scene()).focusRect.refresh()

    def hide(self) ->None:
        """remove all uiTile references so they can be garbage collected"""
        for uiTile in self.uiTiles:
            uiTile.hide()
        self.uiTiles = []
        self._focusTile = None
        if isAlive(self):
            for uiTile in self:
                uiTile.hide()
            super().hide()

    def clear(self) ->None:
        """clear away all tiles"""
        if not self.empty:
            for uiTile in self:
                uiTile.setBoard(None)
            self.uiTiles = []

    def autoSelectTile(self) ->None:
        """call this when Kajongg should automatically focus
        on an appropriate uiTile"""
        focusCandidates = self._focusableTiles()
        if focusCandidates:
            firstCandidate = focusCandidates[0]
            if self._focusTile not in focusCandidates:
                focusCandidates = [x for x in focusCandidates if x.sortKey() >= self.__prevPos]
                focusCandidates.append(firstCandidate)
                self.focusTile = focusCandidates[0]

    @property
    def currentFocusTile(self) ->Optional[UITile]:
        """get focusTile without selecting one"""
        return self._focusTile

    @property
    def focusTile(self) ->Optional[UITile]:
        """the uiTile of this board with focus. This is per Board!"""
        if self._focusTile is None:
            self.autoSelectTile()
        return self._focusTile

    @focusTile.setter
    def focusTile(self, uiTile:Optional[UITile]) ->None:
        """the uiTile of this board with focus. This is per Board!"""
        if uiTile is self._focusTile:
            if Debug.focusable:
                logDebug(f'{self}.focusTile: is already on {uiTile}')
            return
        if uiTile:
            assert uiTile.isKnown, uiTile
            if not isinstance(uiTile.board, DiscardBoard):
                assert uiTile.focusable, uiTile
            self.__prevPos = uiTile.sortKey()
        self._focusTile = uiTile
        if self._focusTile and self._focusTile.tile.exposed.name2() in Debug.focusable:
            logDebug(f"{self.debug_name()}: new focus "
                     f"{self._focusTile if self._focusTile else 'None'} from {stack('')[-1]}")
        if self.hasLogicalFocus:
            scene = self.scene()
            if scene:
                cast('SceneWithFocusRect', scene).focusBoard = self

    def setEnabled(self, enabled:bool) ->None:
        """enable/disable this board"""
        self.tileDragEnabled = enabled
        super().setEnabled(enabled)

    def _focusableTiles(self, sortDir:Qt.Key=Qt.Key.Key_Right) ->List[UITile]:
        """return a list of all tiles in this board sorted such that
        moving in the sortDir direction corresponds to going to
        the next list element.
        respect board orientation: Right Arrow should always move right
        relative to the screen, not relative to the board"""
        return sorted((x for x in self if x.focusable), key=lambda x: x.sortKey(sortDir))

    @property
    def hasLogicalFocus(self) ->bool:
        """defines if this board should show a focusRect
        if another board has focus, setting this to False does
        not change scene.focusBoard

        Up to May 2021, this was called hasFocus, overriding QGraphicsItem.hasFocus
        but pylint did not like that."""
        scene = cast('SceneWithFocusRect', self.scene())
        if scene:
            return scene.focusBoard == self and bool(self._focusTile)
        return False

    @hasLogicalFocus.setter
    def hasLogicalFocus(self, value:bool) ->None:
        """set focus on this board"""
        if isAlive(self):
            scene = cast('SceneWithFocusRect', self.scene())
            if isAlive(scene):
                if scene.focusBoard == self or value:
                    if self.focusTile:
                        assert self.focusTile.board == self, f'{self.focusTile} not in self {self}'
                if value:
                    scene.focusBoard = self

    @staticmethod
    def mapChar2Arrow(event:'QKeyEvent') ->Qt.Key:
        """map the keys hjkl to arrows like in vi and konqueror"""
        key = cast(Qt.Key, event.key())
        if key in Board.arrows:
            return key
        charArrows:str = i18nc(
            'kajongg:arrow keys hjkl like in konqueror',
            'hjklHJKL')
        keychar = event.text()
        if keychar and keychar in charArrows:
            key = Board.arrows[charArrows.index(keychar) % 4]
        return key

    def keyPressEvent(self, event:Optional['QKeyEvent']) ->None:
        """navigate in the board"""
        if event:
            key = Board.mapChar2Arrow(event)
            if key in Board.arrows:
                self.__moveCursor(key)
            else:
                super().keyPressEvent(event)

    def __moveCursor(self, key:Qt.Key) ->None:
        """move focus"""
        tiles = self._focusableTiles(key)
        if tiles:
            # sometimes the handBoard still has focus but
            # has no focusable tiles. Like after declaring
            # Original Call.
            assert self.focusTile
            oldPos = self.focusTile.xoffset, self.focusTile.yoffset
            tiles = [
                x for x in tiles if (x.xoffset,
                                     x.yoffset) != oldPos or x == self.focusTile]
            assert tiles, [str(x) for x in self]
            tiles.append(tiles[0])
            self.focusTile = tiles[tiles.index(self.focusTile) + 1]

    def mapMouseTile(self, uiTile:UITile) ->UITile:
        """map the pressed tile to the wanted tile. Scoring games need this"""
        return uiTile

    def uiMeldWithTile(self, uiTile:UITile, remove:bool=False) ->UIMeld:  # pylint: disable=unused-argument
        """return the UI Meld with uiTile. A Board does not know about melds,
        so default is to return a Meld with only uiTile"""
        return UIMeld(uiTile)

    def meldVariants(self, tile:UITile, forLowerHalf:bool) ->MeldList:  # pylint: disable=unused-argument
        """all possible melds that could be meant by dragging/dropping uiTile"""
        return MeldList(Meld(tile))

    def chooseVariant(self, uiTile:UITile, forLowerHalf:bool=False) ->Optional[Meld]:
        """make the user choose from a list of possible melds for the target.
        The melds do not contain real Tiles, just the scoring strings."""
        variants = self.meldVariants(uiTile, forLowerHalf)
        idx = 0
        if len(variants) > 1:
            menu = QMenu(i18n('Choose from'))
            for idx, variant in enumerate(variants):
                action = menu.addAction(variant.typeName())
                action.setData(idx)
            assert Internal.scene
            if Internal.scene.mainWindow.centralView.dragObject:
                menuPoint = QCursor.pos()
            else:
                assert uiTile.board
                _ = uiTile.board.tileFaceRect().bottomRight()
                view = Internal.scene.mainWindow.centralView
                menuPoint = view.mapToGlobal(
                    view.mapFromScene(cast(QPointF, uiTile.mapToScene(_))))
            action = menu.exec(menuPoint)
            if not action:
                return None
            idx = action.data()
        return variants[idx]

    def dragEnterEvent(self, unusedEvent:Optional['QGraphicsSceneDragDropEvent']) ->None:
        """drag enters the HandBoard: highlight it"""
        self.setPen(QPen(self.penColor))

    def dragLeaveEvent(self, unusedEvent:Optional['QGraphicsSceneDragDropEvent']) ->None:
        """drag leaves the HandBoard"""
        self._noPen()

    def _noPen(self) ->None:
        """remove pen for this board. The pen defines the border"""
        if Debug.graphics:
            self.setPen(QPen(self.penColor))
        else:
            self.setPen(QPen(Qt.PenStyle.NoPen))

    def tilesByElement(self, element:Tile) ->List[UITile]:
        """return all child items holding a uiTile for element"""
        return [x for x in self if x.tile is element]

    def rotatedLightSource(self) ->str:
        """the light source we need for the original uiTile before it is rotated"""
        lightSourceIndex = LIGHTSOURCES.index(self.lightSource)
        lightSourceIndex = (lightSourceIndex + sceneRotation(self) // 90) % 4
        return LIGHTSOURCES[lightSourceIndex]

    def tileFacePos(self, showShadows:Optional[bool]=None) ->QPointF:
        """the face pos of a uiTile relative to its origin"""
        if showShadows is None:
            assert Internal.Preferences
            showShadows = bool(Internal.Preferences.showShadows)
        if not showShadows:
            return QPointF()
        lightSource = self.rotatedLightSource()
        xoffset = self.tileset.shadowWidth() - 1 if 'E' in lightSource else 0
        yoffset = self.tileset.shadowHeight() - 1 if 'S' in lightSource else 0
        return QPointF(xoffset, yoffset)

    def tileFaceRect(self) ->QRectF:
        """the face rect of a uiTile relative its origin"""
        return QRectF(self.tileFacePos(), self.tileset.faceSize)

    def setTilePos(self, xWidth:int=0, xHeight:float=0.0, yWidth:int=0, yHeight:float=0.0) ->None:
        """set the position in the parent item expressing the position in tile face units.
        The X position is xWidth*facewidth + xHeight*faceheight, analog for Y"""
        self.__xWidth = xWidth
        self.__xHeight = xHeight
        self.__yWidth = yWidth
        self.__yHeight = yHeight
        self.setGeometry()

    def setBoardRect(self, width:float, height:float) ->None:
        """gives the board a fixed size in uiTile coordinates"""
        self.__fixedWidth = width
        self.__fixedHeight = height
        self.computeRect()

    def computeRect(self) ->None:
        """translate from our rect coordinates to scene coord"""
        sizeX = self.tileset.faceSize.width() * self.__fixedWidth
        sizeY = self.tileset.faceSize.height() * self.__fixedHeight
        rect = self.rect()
        if Internal.Preferences:
            if Internal.Preferences.showShadows:
                sizeX += self.tileset.shadowWidth()
                sizeY += self.tileset.shadowHeight()
        rect.setWidth(sizeX)
        rect.setHeight(sizeY)
        self.prepareGeometryChange()
        super().setRect(rect)

    @property
    def width(self) ->float:
        """getter for width"""
        return self.__fixedWidth

    @property
    def height(self) ->float:
        """getter for width"""
        return self.__fixedHeight

    def setGeometry(self) ->None:
        """move the board to the correct position and set its rect surrounding all its
        items. This is needed for enabling drops into the board.
        This is also called when the tileset or the light source for this board changes"""
        width = self.tileset.faceSize.width()
        height = self.tileset.faceSize.height()
        assert Internal.Preferences
        if not Internal.Preferences.showShadows:
            offsets = (0, 0)
        elif self.isHandBoard:
            offsets = (-self.tileset.shadowHeight() * 2, 0)
        else:
            offsets = self.tileset.shadowOffsets(
                self._lightSource,
                sceneRotation(self))
        newX = self.__xWidth * width + self.__xHeight * height + offsets[0]
        newY = self.__yWidth * width + self.__yHeight * height + offsets[1]
        self.setPos(newX, newY)

    def showShadowsChanged(self, unusedOldValue:bool, newValue:bool) ->None:
        """set active lightSource"""
        for uiTile in self:
            uiTile.setClippingFlags()
        self._reload(self.tileset, showShadows=newValue)

    @property
    def lightSource(self) ->str:
        """the active lightSource"""
        return self._lightSource

    @lightSource.setter
    def lightSource(self, lightSource:str) ->None:
        """set active lightSource"""
        if self._lightSource != lightSource:
            if lightSource not in LIGHTSOURCES:
                logException(f'lightSource {lightSource} illegal')
            self._reload(self.tileset, lightSource)

    @property
    def tileset(self) ->Tileset:
        """get/set the active tileset and resize accordingly"""
        return self._tileset

    @tileset.setter
    def tileset(self, tileset:Tileset) ->None:
        """get/set the active tileset and resize accordingly"""
        self._reload(tileset, self._lightSource)

    def _reload(self, tileset:Optional[Tileset]=None,
        lightSource:Optional[str]=None, showShadows:Optional[bool]=None) ->None:
        """call this if tileset or lightsource change: recomputes the entire board"""
        assert Internal.Preferences
        if tileset is None:
            tileset = self.tileset
        if lightSource is None:
            lightSource = self._lightSource
        if showShadows is None:
            showShadows = bool(Internal.Preferences.showShadows)
        if (self._tileset != tileset
                or self.__yHeight == 0.0
                or self._lightSource != lightSource
                or Internal.Preferences.showShadows != showShadows):
            self.prepareGeometryChange()
            self._tileset = tileset
            self._lightSource = lightSource
            self.setGeometry()
            for uiTile in self:
                self.placeTile(uiTile)
                uiTile.update()
            self.computeRect()
            if self.hasLogicalFocus:
                cast('SceneWithFocusRect', self.scene()).focusBoard = self

    def focusRectWidth(self) ->int:
        """how many tiles are in focus rect?"""
        return 1

    def shiftZ(self, level:int) ->QPointF:
        """used for 3D: compute the needed shift for the uiTile.
        level is the vertical position. 0 is the face position on
        ground level, -1 is the imprint a uiTile makes on the
        surface it stands on"""
        assert Internal.Preferences
        if not Internal.Preferences.showShadows:
            return QPointF()
        shiftX = 0.0
        shiftY = 0.0
        if level != 0:
            lightSource = self.rotatedLightSource()
            stepX = level * self.tileset.shadowWidth() / 2
            stepY = level * self.tileset.shadowHeight() / 2
            if 'E' in lightSource:
                shiftX = stepX
            if 'W' in lightSource:
                shiftX = -stepX
            if 'N' in lightSource:
                shiftY = -stepY
            if 'S' in lightSource:
                shiftY = stepY
        return QPointF(shiftX, shiftY)

    def tileSize(self) ->'QSizeF':
        """the current uiTile size"""
        return self._tileset.tileSize

    def faceSize(self) ->'QSizeF':
        """the current face size"""
        return self._tileset.faceSize

    @overload
    def __getitem__(self, index:Union[int, str, Tile]) -> UITile:
        ...
    @overload
    def __getitem__(self, index:slice) -> List[UITile]:
        ...

    def __getitem__(self, index):
        """board[-1] would return rightmost tile, sorted by X position"""
        if isinstance(index, int):
            return self.uiTiles[index]
        if isinstance(index, Tile):
            return sorted((x for x in self if x.tile is index), key=lambda x: x.sortKey())
        if isinstance(index, str):
            _ = Tile(index)
            return sorted((x for x in self if x.tile is _), key=lambda x: x.sortKey())
        if isinstance(index, slice):
            return sorted(self.uiTiles.__getitem__(index), key=lambda x: x.sortKey())
        raise NotImplementedError(f': index={index}')

    def __iter__(self) -> Iterator[UITile]:
        """just to make this clear to mypy"""
        return iter(self.uiTiles)

    def __len__(self) ->int:
        """len(Board) returns # of UITiles"""
        return len(self.uiTiles)

    def __bool__(self) ->bool:
        """because a Board with len == 0 would return False"""
        return True

    def placeTile(self, uiTile:UITile) ->None:
        """places the uiTile in the scene"""
        assert isinstance(uiTile, UITile)
        assert uiTile.board == self
        uiTile.setupAnimations()

    def addUITile(self, uiTile:UITile) ->None:
        """add uiTile to this board"""
        if Debug.uitiles:
            logDebug(f'{self.debug_name()}.addUITile({uiTile})')
        self.uiTiles.append(uiTile)

    def removeUITile(self, uiTile:UITile) ->None:
        """remove uiTile from this board"""
        if Debug.uitiles:
            logDebug(f'{self.debug_name()}.removeUITile({uiTile})')
        self.uiTiles.remove(uiTile)
        if self.currentFocusTile == uiTile:
            self.focusTile = None


class CourtBoard(Board):

    """A Board that is displayed within the wall"""
    penColor = QColor('green')

    def __init__(self, width:int, height:int) ->None:
        super().__init__(width, height, Tileset.current())
        self.setAcceptDrops(True)

    def maximize(self) ->None:
        """make it as big as possible within the wall"""
        assert Internal.scene
        assert Internal.scene.game
        cWall:'UIWall' = cast('UIWall', Internal.scene.game.wall)
        if not cWall:
            return
        newSceneX = cWall[3].sceneBoundingRect().right()
        newSceneY = cWall[2].sceneBoundingRect().bottom()
        tileset = self.tileset
        xAvail = cWall[1].sceneBoundingRect().left() - newSceneX
        yAvail = cWall[0].sceneBoundingRect().top() - newSceneY
        shadowHeight = tileset.shadowHeight()
        shadowWidth = tileset.shadowWidth()
        assert Internal.Preferences
        if Internal.Preferences.showShadows:
            # this should use the real shadow values from the wall because the wall
            # tiles are smaller than those in the CourtBoard but this should be
            # good enough
            newSceneX -= shadowHeight / 2 if 'W' in self.lightSource else 0
            newSceneY -= shadowWidth / 2 if 'N' in self.lightSource else 0
            xAvail -= shadowHeight if 'E' in self.lightSource else 0
            yAvail -= shadowWidth if 'S' in self.lightSource else 0
        xNeeded = self.width * tileset.faceSize.width()
        yNeeded = self.height * tileset.faceSize.height()
        xScaleFactor = xAvail / xNeeded
        yScaleFactor = yAvail / yNeeded
        self.setPos(newSceneX, newSceneY)
        self.setScale(min(xScaleFactor, yScaleFactor))
        for uiTile in self:
            assert uiTile.board
            uiTile.board.placeTile(uiTile)


class SelectorBoard(CourtBoard):

    """a board containing all possible tiles for selection"""

    def __init__(self) ->None:
        super().__init__(9, 7)
        self.allSelectorTiles:List[UITile] = []

    def checkTiles(self) ->None:
        """does not apply"""

    def load(self, game:'Game') ->None:
        """load the tiles according to game.ruleset"""
        self.clear()
        self.allSelectorTiles = [UITile(x) for x in elements.all(game.ruleset)]
        self.refill()

    def refill(self) ->None:
        """move all tiles back into the selector"""
        with AnimationSpeed():
            for uiTile in self.allSelectorTiles:
                uiTile.change_name(uiTile.exposed)
                self.__placeAvailable(uiTile)
                uiTile.dark = False
                uiTile.focusable = True
            self.focusTile = self['c1'][0]  # type:ignore[index]

    def debug_name(self) ->str:
        """for debugging messages"""
        return 'selector'

    def dragMoveEvent(self, event:Optional['QGraphicsSceneDragDropEvent']) ->None:
        """allow dropping only from handboards, not from self"""
        if event:
            uiTile = cast(MimeData, event.mimeData()).uiTile
            event.setAccepted(uiTile.board != self)

    def dropEvent(self, event:Optional['QGraphicsSceneDragDropEvent']) ->None:
        """drop a uiTile into the selector"""
        if event:
            uiTile = cast(MimeData, event.mimeData()).uiTile
            self.dropMeld(uiTile)
            event.accept()

    def dropMeld(self, uiTile:UITile, forLowerHalf:bool=False) ->None:  # pylint: disable=unused-argument
        """drop UIMeld containing uiTile into selector board"""
        assert uiTile.board
        senderHand = uiTile.board
        uiMeld = senderHand.uiMeldWithTile(uiTile, remove=True)
        if senderHand == self:
            return
        for myTile in uiMeld:
            self.__placeAvailable(myTile)
            myTile.focusable = True
        (senderHand if not senderHand.empty else self).hasLogicalFocus = True
        self._noPen()
        animate()
        cast('ScoringScene', Internal.scene).handSelectorChanged(cast('HandBoard', senderHand))

    def loseMeld(self, uiTile:UITile, meld:Meld) ->UIMeld:
        """loses a UIMeld. First uiTile is given, the rest should be as defined by meld"""
        assert isinstance(uiTile, UITile), uiTile
        result = UIMeld(uiTile)
        for tile in meld[1:]:
            baseTiles = [
                x for x in self.tilesByElement(
                    tile.exposed) if x not in result]
            result.append(baseTiles[0])
        return result

    def __placeAvailable(self, uiTile:UITile) ->None:
        """place the uiTile in the selector at its place"""
        # define coordinates and order for tiles:
        offsets = {
            Tile.dragon: (3, 6, Tile.dragons),
            Tile.flower: (4, 5, Tile.winds),
            Tile.season: (4, 0, Tile.winds),
            Tile.wind: (3, 0, Tile.winds),
            Tile.bamboo: (1, 0, '123456789'),
            Tile.stone: (2, 0, '123456789'),
            Tile.character: (0, 0, '123456789')}
        row, baseColumn, order = offsets[uiTile.lowerGroup]
        column = baseColumn + order.index(uiTile.char)
        uiTile.dark = False
        uiTile.setBoard(self, column, row)

    def meldVariants(self, tile:UITile, forLowerHalf:bool) ->MeldList:
        """return a list of possible variants based on meld. Those are logical melds."""
        assert isinstance(tile, UITile)
        wantedTile = tile.tile
        for selectorTile in self.uiTiles:
            selectorTile.change_name(selectorTile.exposed)
        lowerName = wantedTile.exposed
        upperName = wantedTile.concealed
        if forLowerHalf:
            scName = upperName
        else:
            scName = lowerName
        result = MeldList(scName.single)
        baseTiles = len(self.tilesByElement(lowerName))
        if baseTiles >= 2:
            result.append(scName.pair)
        if baseTiles >= 3:
            result.append(scName.pung)
        if baseTiles == 4:
            if forLowerHalf:
                result.append(lowerName.kong.declared)
            else:
                result.append(lowerName.kong)
                result.append(lowerName.kong.exposedClaimed)
        if wantedTile.isNumber and wantedTile.value < 8:  # type:ignore[operator]
            chow2 = scName.nextForChow
            chow3 = chow2.nextForChow
            if self.tilesByElement(chow2.exposed) and self.tilesByElement(chow3.exposed):
                result.append(scName.chow)
        # result now holds a list of melds
        return result


class MimeData(QMimeData):

    """we also want to pass a reference to the moved meld"""

    def __init__(self, uiTile:UITile) ->None:
        QMimeData.__init__(self)
        self.uiTile:UITile = uiTile
        self.setText(str(uiTile))


class FittingView(QGraphicsView):

    """a graphics view that always makes sure the whole scene is visible"""

    def __init__(self, parent:Optional['QWidget']=None) ->None:
        """generate a fitting view with our favourite properties"""
        super().__init__(parent)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        vpol = QSizePolicy()
        vpol.setHorizontalPolicy(QSizePolicy.Policy.Expanding)
        vpol.setVerticalPolicy(QSizePolicy.Policy.Expanding)
        self.setSizePolicy(vpol)
        self.setRenderHint(QPainter.RenderHint.Antialiasing)
        self.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform)
        self.setStyleSheet('background: transparent')
        self.setFrameStyle(QFrame.Shape.NoFrame)
        self.tilePressed:Optional[UITile] = None
        self.dragObject:Optional[QDrag] = None
        self.setFocus()

    def wheelEvent(self, event:Optional['QWheelEvent']) ->None:
        """we do not want scrolling for the scene view.
        Instead scrolling down changes perspective like in kmahjongg"""
        if event:
            angleX = event.angleDelta().x()
            angleY = event.angleDelta().y()
            if angleX > 15 or angleY > -50:
                return
            assert Internal.mainWindow
            Internal.mainWindow.changeAngle()

    def resizeEvent(self, unusedEvent:Optional['QResizeEvent']) ->None:
        """scale the scene and its background for new view size"""
        assert Internal.Preferences
        Internal.Preferences.callTrigger(
            'tilesetName')  # this redraws and resizes
        Internal.Preferences.callTrigger('backgroundName')  # redraw background
        if scene := self.scene():
            self.fitInView(
                scene.itemsBoundingRect(),
                Qt.AspectRatioMode.KeepAspectRatio)
        self.setFocus()

    def __matchingTile(self, position:QPoint, uiTile:UITile) ->bool:
        """is position in the clickableRect of this uiTile?"""
        if not isinstance(uiTile, UITile):
            return False
        itemPos = uiTile.mapFromScene(self.mapToScene(position))
        assert uiTile.board
        return uiTile.board.tileFaceRect().contains(itemPos)

    def tileAt(self, position:QPoint) ->List[UITile]:
        """find out which uiTile is clickable at this position. Always
        returns a list. If there are several tiles above each other,
        return all of them, highest first"""
        allTiles = [x for x in self.items(position) if isinstance(x, UITile)]
        items = [x for x in allTiles if self.__matchingTile(position, x)]
        if not items:
            return []
        for item in items[:]:
            for other in allTiles:
                if (other.xoffset, other.yoffset) == (item.xoffset, item.yoffset):
                    if other.level > item.level:
                        items.append(other)
        return uniqueList(sorted(items, key=lambda x: -x.level))

    def __setFocusTile(self, event:'QMouseEvent') ->Optional[UITile]:
        """set focus on first tile of meld"""
        tiles = self.tileAt(event.pos())
        if tiles:
            tile = tiles[0]
            board = tile.board
            assert board
            uiTile = board.mapMouseTile(tile)
            if uiTile.focusable:
                board.focusTile = uiTile
                board.hasLogicalFocus = True
                assert Internal.scene
                if hasattr(Internal.scene, 'clientDialog'):
                    if Internal.scene.clientDialog:
                        Internal.scene.clientDialog.buttons[0].setFocus()
                return uiTile
        event.ignore()
        return None

    def mousePressEvent(self, event:Optional['QMouseEvent']) ->None:
        """set blue focus frame"""
        if event:
            self.tilePressed = self.__setFocusTile(event)
            if event.modifiers() & Qt.KeyboardModifier.ShiftModifier:
                for uiTile in self.tileAt(event.pos()):
                    assert uiTile.board
                    print(f'{uiTile}: board.level:{uiTile.board.level}')
            super().mousePressEvent(event)

    def mouseReleaseEvent(self, event:Optional['QMouseEvent']) ->None:
        """release self.tilePressed"""
        if event:
            super().mouseReleaseEvent(event)
            # scene.focusItem() may have disappeared since mousePressEvent:
            # 1. start scoring game, if needed select main window by clicking on its frame
            # 2. click on a tile in the center. This will set focusItem to None!
            # 3. use cursor to move position (hjkl): this fails
            # so we must call __setFocusTile() again
            self.__setFocusTile(event)
            self.tilePressed = None

    def mouseMoveEvent(self, event:Optional['QMouseEvent']) ->None:
        """selects the correct uiTile"""
        if event:
            tilePressed = self.tilePressed
            self.tilePressed = None
            if tilePressed:
                board = tilePressed.board
                if board and board.tileDragEnabled:
                    self.dragObject = self.drag(tilePressed)
                    self.dragObject.exec(Qt.DropAction.MoveAction)
                    self.dragObject = None
                    return
            super().mouseMoveEvent(event)

    def drag(self, uiTile:UITile) ->QDrag:
        """return a drag object"""
        drag = QDrag(self)
        mimeData = MimeData(uiTile)
        drag.setMimeData(mimeData)
        tRect = uiTile.boundingRect()
        tRect = self.viewportTransform().mapRect(tRect)
        pmapSize = QSize(
            int(tRect.width() * cast(float, uiTile.scale)),
            int(tRect.height() * cast(float, uiTile.scale)))
        pMap = uiTile.pixmapFromSvg(pmapSize)
        drag.setPixmap(pMap)
        drag.setHotSpot(QPoint(pMap.width() // 2, pMap.height() // 2))
        return drag


class YellowText(QGraphicsRectItem):

    """a yellow rect with a message, used for claims"""

    def __init__(self, parent:QGraphicsRectItem) ->None:
        super().__init__(parent)
        self.parent = parent
        self.font:QFont = QFont()
        self.font.setPointSize(48)
        self.height:int = 62
        self.width:int = 200
        self.msg:Optional[str] = None
        self.setText('')

    def setText(self, msg:str) ->None:
        """set the text of self"""
        self.msg = f'{msg}  '
        metrics = QFontMetrics(self.font)
        self.width = metrics.horizontalAdvance(self.msg)
        self.height = int(metrics.lineSpacing() * 1.1)
        self.setRect(0, 0, self.width, self.height)
        self.resetTransform()
        self.setPos(self.parent.rect().center())
        rotation = self.parent.rotation()
        rotateCenter(self, -rotation)
        xOffset = -self.rect().width() / 2
        yOffset = -self.rect().height() / 2
        if rotation % 180 == 0:
            self.moveBy(xOffset, yOffset * 4)
        else:
            self.moveBy(xOffset, yOffset)

    def paint(self, painter:Optional[QPainter], unusedOption:Optional[QStyleOptionGraphicsItem],
        unusedWidget:Optional['QWidget']=None) ->None:
        """override predefined paint"""
        if painter:
            painter.setFont(self.font)
            painter.fillRect(self.rect(), QBrush(QColor('yellow')))
            if self.msg:
                painter.drawText(self.rect(), self.msg)


class DiscardBoard(CourtBoard):

    """A special board for discarded tiles"""
    penColor = QColor('orange')

    def __init__(self) ->None:
        super().__init__(11, 9)
        self.__places:List[Tuple[float, int]]
        self.__lastDiscarded:Optional[UITile] = None
        self.__discardTilesOrderedLeaveHole:bool = True

    def debug_name(self) ->str:
        """to be used in debug output"""
        return "discardBoard"

    def hide(self) ->None:
        """remove all uiTile references so they can be garbage collected"""
        self.__lastDiscarded = None
        super().hide()

    def setRandomPlaces(self, game:'PlayingGame') ->None:
        """precompute random positions"""
        assert isinstance(self.width, int)
        assert isinstance(self.height, int)
        self.__places = [(x, y) for x in range(self.width)
                         for y in range(self.height)]
        if game.ruleset.discardTilesOrdered:
            self.__places.sort(key=lambda p: p[0] + p[1] * 1000)
        else:
            game.randomGenerator.shuffle(self.__places)
        self.__discardTilesOrderedLeaveHole = game.ruleset.discardTilesOrderedLeaveHole

    def discardTile(self, uiTile:UITile) ->None:
        """add uiTile to the discard board"""
        assert isinstance(uiTile, UITile)
        uiTile.setBoard(self, *self.__places.pop(0))
        uiTile.dark = False
        uiTile.focusable = False
        self.focusTile = uiTile
        self.hasLogicalFocus = True
        self.__lastDiscarded = uiTile

    def claimDiscard(self) ->UITile:
        """claim last discarded tile"""
        result = self.__lastDiscarded
        assert result
        self.__lastDiscarded = None
        if not self.__discardTilesOrderedLeaveHole:
            self.__places.insert(0, (result.xoffset, result.yoffset))
        return result

    def dropEvent(self, event:Optional['QGraphicsSceneDragDropEvent']) ->None:
        """drop a uiTile into the discard board

        The user uses the mouse for discarding a tile"""
        if event:
            assert Internal.scene
            uiTile = cast(MimeData, event.mimeData()).uiTile
            assert isinstance(uiTile, UITile), uiTile
            _ = event.scenePos()
            _ -= uiTile.boundingRect().center()
            uiTile.setPos(_)
            if Internal.scene.clientDialog:
                Internal.scene.clientDialog.selectButton(cast('ClientMessage', Message.Discard))
            event.accept()
        self._noPen()



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

partially based on C++ code from:
 - Copyright (C) 2006 Mauricio Piacentini <mauricio@tabuleiro.com>

SPDX-License-Identifier: GPL-2.0-only

"""

from collections import defaultdict
from typing import Dict, Optional, Union, TYPE_CHECKING, Callable, Any, cast

from kde import KConfigSkeleton
from log import logException, logDebug
from common import Internal, Debug

if TYPE_CHECKING:
    from kdestub import KConfigSkeletonItem, ItemInt


class Parameter:

    """helper class for defining configuration parameters"""

    def __init__(self, group:str, name:str, default:Union[int, str, bool]) ->None:
        """configuration group, parameter name, default value"""
        self.group = group
        self.name = name
        self.default = default
        self.item:'KConfigSkeletonItem'

    def add(self, skeleton:KConfigSkeleton) ->None:
        """for mypy"""
        raise NotImplementedError

    def itemValue(self) ->Union[int, str, bool]:
        """return the value of this item"""
        return self.item.value()


class StringParameter(Parameter):

    """helper class for defining string parameters"""

    def __init__(self, group:str, name:str, default:Optional[str]=None) ->None:
        if default is None:
            default = ''
        super().__init__(group, name, default)
        self.value = ''

    def add(self, skeleton:KConfigSkeleton) ->None:
        """add tis parameter to the skeleton"""
        self.item = skeleton.addItem(self.name, self.value, self.default or '')

    def itemValue(self) ->str:
        """return the value of this item"""
        return str(self.item.value())


class BoolParameter(Parameter):

    """helper class for defining boolean parameters"""

    def __init__(self, group:str, name:str, default:Optional[bool]=None) ->None:
        if default is None:
            default = False
        super().__init__(group, name, default)
        self.value = default

    def add(self, skeleton:KConfigSkeleton) ->None:
        """add tis parameter to the skeleton"""
        self.item = skeleton.addItem(self.name, self.value, self.default)


class IntParameter(Parameter):

    """helper class for defining integer parameters"""

    def __init__(self, group:str, name:str, default:Optional[int]=None,
                 minValue:Optional[int]=None, maxValue:Optional[int]=None) ->None:
        """minValue and maxValue are also used by the edit widget"""
        if default is None:
            default = 0
        super().__init__(group, name, default)
        self.value = default
        self.minValue = minValue
        self.maxValue = maxValue

    def add(self, skeleton:KConfigSkeleton) ->None:
        """add this parameter to the skeleton"""
        self.item = skeleton.addItem(self.name, self.value, self.default)
        if self.minValue is not None:
            cast('ItemInt', self.item).setMinValue(self.minValue)
        if self.maxValue is not None:
            cast('ItemInt', self.item).setMaxValue(self.maxValue)


class SetupPreferences(KConfigSkeleton):

    """Holds all Kajongg options. Only instantiate this once"""
    _Parameters : Dict[str, Parameter] = {}

    def __init__(self) ->None:
        if Internal.Preferences:
            logException('Preferences is not None')
        self.__watching = defaultdict(list)  # type:ignore
        Internal.Preferences = self
        super().__init__()
        self.configChanged.connect(self.callTriggers)
        self.__oldValues  = defaultdict(str)  # type:ignore
        self.addString('General', 'tilesetName', 'default')
        self.addString('General', 'windTilesetName', 'traditional')
        self.addString('General', 'backgroundName', 'wood_light')
        self.addBool('Display', 'showShadows', True)
        self.addBool('Display', 'rearrangeMelds', False)
        self.addBool('Display', 'showOnlyPossibleActions', True)
        self.addBool('Display', 'propose', True)
        self.addInteger('Display', 'animationSpeed', 70, 0, 99)
        self.addBool('Display', 'useSounds', True)
        self.addBool('Display', 'uploadVoice', False)

    def callTrigger(self, name:str) ->None:
        """call registered callback for this attribute change"""
        newValue = getattr(self, name)
        if self.__oldValues[name] != newValue:
            if Debug.preferences:
                logDebug(f"{name}: {self.__oldValues[name]} -> "
                         f"{newValue} calling {','.join(x.__name__ for x in self.__watching[name])}")
            for method in self.__watching[name]:
                method(self.__oldValues[name], newValue)
        self.__oldValues[name] = newValue

    def callTriggers(self) ->None:
        """call registered callbacks for specific attribute changes"""
        for name in self.__watching:
            self.callTrigger(name)

    #def addWatch(self, name:str, method:Callable[List[Union[int, str, bool]], None]) ->None:
    def addWatch(self, name:str, method:Callable[..., None]) ->None:
        """If name changes, call method.
        method must accept 2 arguments: old value and new value."""
        if method not in self.__watching[name]:
            self.__watching[name].append(method)
            if Debug.preferences:
                logDebug(f'addWatch on {name}, hook {method.__name__}')
            self.callTrigger(name)  # initial change

    def __getattr__(self, name:str) ->Union[int, str, bool]:
        """undefined attributes might be parameters"""
        if name not in SetupPreferences._Parameters:
            return self.__getattribute__(name)
        par = SetupPreferences._Parameters[name]
        return par.itemValue()

    def __setattr__(self, name:str, value:Union[int, str, bool]) ->None:
        """undefined attributes might be parameters"""
        if name not in SetupPreferences._Parameters:
            super().__setattr__(name, value)
            return
        par = SetupPreferences._Parameters[name]
        par.item.setValue(value)

    def __getitem__(self, key:str):  # type:ignore
        return self.__getattr__(key)

    def __setitem__(self, key:str, value:Any) ->None:
        self.__setattr__(key, value)

    def __delitem__(self, key:str) ->None:
        """pylint wants this for a complete container, but we do not need it"""
        del SetupPreferences._Parameters[key]

    def __len__(self) ->int:
        """pylint wants this for a complete container, but we do not need it"""
        return len(SetupPreferences._Parameters)

    def addParameter(self, par:Parameter) ->None:
        """add a parameter to the skeleton"""
        if par.name not in SetupPreferences._Parameters:
            SetupPreferences._Parameters[par.name] = par
            self.setCurrentGroup(par.group)
            par.add(self)

    def addString(self, group:str, name:str, default:Optional[str]=None) ->None:
        """add a string parameter to the skeleton"""
        self.addParameter(StringParameter(group, name, default))

    def addBool(self, group:str, name:str, default:Optional[bool]=None) ->None:
        """add a string parameter to the skeleton"""
        self.addParameter(BoolParameter(group, name, default))

    def addInteger(self, group:str, name:str, default:Optional[int]=None,
                   minValue:Optional[int]=None, maxValue:Optional[int]=None) ->None:
        """add a string parameter to the skeleton"""
        self.addParameter(IntParameter(
            group, name, default, minValue, maxValue))

    def animationDuration(self) ->int:
        """in milliseconds"""
        return max(0, (99 - self.animationSpeed) * 100 // 4)  # type:ignore



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import List, Tuple, TYPE_CHECKING, Optional, Union

from qt import Qt, QAbstractTableModel, QModelIndex, QPersistentModelIndex
from qt import QLabel, QDialog, QHBoxLayout, QVBoxLayout, QDialogButtonBox

from mi18n import i18n, i18nc
from statesaver import StateSaver
from guiutil import ListComboBox, MJTableView, decorateWindow
from guiutil import BlockSignals
from common import Debug
from modeltest import ModelTest

if TYPE_CHECKING:
    from qt import QWidget
    from rule import Ruleset


class DifferModel(QAbstractTableModel):

    """a model for our ruleset differ"""

    def __init__(self, diffs:List[Tuple[str,str,str]], view:'RulesetDiffer') ->None:
        super().__init__()
        self.diffs = diffs
        self.view = view

    def columnCount(self, unusedIndex:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->int:
        """how many columns does this node have?"""
        return 3  # rule name, left values, right values

    def rowCount(self, parent:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->int:
        """how many items?"""
        if parent.isValid():
            # we have only top level items
            return 0
        return len(self.diffs)

    def data(self, index:Union[QModelIndex,QPersistentModelIndex],
        role:int=Qt.ItemDataRole.DisplayRole) ->Union[int,str,None]:
        """get from model"""
        if not index.isValid():
            return None
        if not 0 <= index.row() < len(self.diffs):
            return None
        diff = self.diffs[index.row()]
        column = index.column()
        if role == Qt.ItemDataRole.DisplayRole:
            return diff[column]
        if role == Qt.ItemDataRole.TextAlignmentRole:
            return int(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)
        return None

    def headerData(self, section:int, orientation:Qt.Orientation,
        role:int=Qt.ItemDataRole.DisplayRole) ->Union[int,str,None]:
        """tell the view about the wanted headers"""
        if role == Qt.ItemDataRole.TextAlignmentRole:
            if orientation == Qt.Orientation.Horizontal:
                return int(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)
        if role != Qt.ItemDataRole.DisplayRole:
            return None
        if orientation == Qt.Orientation.Horizontal:
            if section == 0:
                return i18nc('Kajongg', 'Rule')
            if section == 1:
                return i18n(self.view.cbRuleset1.current.name)
            if section == 2:
                return i18n(self.view.cbRuleset2.current.name)
        return None


class RulesetDiffer(QDialog):

    """Shows differences between rulesets"""

    def __init__(self, leftRulesets:List['Ruleset'], rightRulesets:List['Ruleset'],
        parent:Optional['QWidget']=None) ->None:
        super().__init__(parent)
        leftRulesets, rightRulesets = leftRulesets[:], rightRulesets[:]
        # remove rulesets from right which are also on the left side
        for left in leftRulesets:
            left.load()
        for right in rightRulesets:
            right.load()
        for left in leftRulesets:
            for right in rightRulesets[:]:
                if left == right and left.name == right.name:
                    # rightRulesets.remove(right) this is wrong because it
                    # removes the first ruleset with the same hash
                    rightRulesets = [
                        x for x in rightRulesets if id(x) != id(right)]
        self.leftRulesets = leftRulesets
        self.rightRulesets = rightRulesets
        self.model:DifferModel
        self.modelTest:ModelTest
        self.view = MJTableView(self)
        self.buttonBox = QDialogButtonBox()
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        cbLayout = QHBoxLayout()
        self.cbRuleset1 = ListComboBox(self.leftRulesets)
        if len(self.leftRulesets) == 1:
            self.lblRuleset1 = QLabel(self.leftRulesets[0].name)
            cbLayout.addWidget(self.lblRuleset1)
        else:
            cbLayout.addWidget(self.cbRuleset1)
        self.cbRuleset2 = ListComboBox(self.rightRulesets)
        cbLayout.addWidget(self.cbRuleset2)
        cmdLayout = QHBoxLayout()
        cmdLayout.addWidget(self.buttonBox)
        layout = QVBoxLayout()
        layout.addLayout(cbLayout)
        layout.addWidget(self.view)
        layout.addLayout(cmdLayout)
        self.setLayout(layout)

        decorateWindow(self, i18nc("@title:window", "Compare"))
        self.setObjectName('RulesetDiffer')

        self.cbRuleset1.currentIndexChanged.connect(self.leftRulesetChanged)
        self.cbRuleset2.currentIndexChanged.connect(self.rulesetChanged)
        self.leftRulesetChanged()
        StateSaver(self)

    def leftRulesetChanged(self) ->None:
        """slot to be called if the left ruleset changes"""
        if len(self.leftRulesets) == 1:
            self.orderRight()
        self.rulesetChanged()

    def rulesetChanged(self) ->None:
        """slot to be called if the right ruleset changes"""
        self.model = DifferModel(self.formattedDiffs(), self)
        if Debug.modelTest:
            self.modelTest = ModelTest(self.model, self)
        self.view.setModel(self.model)

    def orderRight(self) ->None:
        """order the right rulesets by similarity to current left ruleset.
        Similarity is defined by the length of the diff list."""
        leftRuleset = self.cbRuleset1.current
        diffPairs = sorted((len(x.diff(leftRuleset)), x)
                           for x in self.rightRulesets)
        combo = self.cbRuleset2
        with BlockSignals([combo]):
            combo.items = [x[1] for x in diffPairs]
        combo.setCurrentIndex(0)

    def formattedDiffs(self) ->List[Tuple[str, str, str]]:
        """a list of tuples with 3 values: name, left, right"""
        formatted = []
        leftRuleset = self.cbRuleset1.current
        rightRuleset = self.cbRuleset2.current
        assert rightRuleset, self.cbRuleset2.count()
        leftRuleset.load()
        rightRuleset.load()
        for rule1, rule2 in leftRuleset.diff(rightRuleset):
            name = i18n(rule1.name if rule1 else rule2.name)
            left = rule1.i18nStr() if rule1 else i18nc(
                'Kajongg-Rule', 'not defined')
            right = rule2.i18nStr() if rule2 else i18nc(
                'Kajongg-Rule', 'not defined')
            formatted.append((name, left, right))
            if rule1 and rule2 and rule1.definition != rule2.definition:
                formatted.append(('', rule1.definition, rule2.definition))
        return formatted



"""
##
## Copyright (C) 2007 Trolltech ASA. All rights reserved.
## Copyright (C) 2010-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>
##
## This file is part of the Qt Concurrent project on Trolltech Labs.
##
## This file may be used under the terms of the GNU General Public
## License version 2.0 as published by the Free Software Foundation
## and appearing in the file LICENSE.GPL included in the packaging of
## this file.  Please review the following information to ensure GNU
## General Public Licensing requirements will be met:
## https://www.qt.io/download-open-source
##
## If you are unsure which license is appropriate for your use, please
## review the following information:
## https://www.qt.io/licensing/ or contact the
## sales department at https://www.qt.io/contact-us/sales-contact-request/?hsLang=en
##
## This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
## WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
##
#############################################################################

There are several variations of python ports for this script floating
around, it seemed easier to me to maintain one for myself
"""

from typing import List, Dict, Any, Type, cast

from qt import QObject, Qt, QAbstractItemModel, QModelIndex, sip_cast
from qt import QPersistentModelIndex, QFont, QColor, QSize
from common import isAlive

# pylint: skip-file

def isValid(variant:Any) ->bool:
    return (variant is not None)


class ModelTest(QObject):

    """tests a model"""

    def __init__(self, _model:QAbstractItemModel, parent:QObject) ->None:
        """
        Connect to all of the models signals, Whenever anything happens recheck everything.
        """
        super().__init__(parent)
        self._model = _model
        self.model = cast(QAbstractItemModel, sip_cast(_model, QAbstractItemModel))  # type:ignore[arg-type]
        self.insert:List[Dict] = []
        self.remove:List[Dict] = []
        self.changing:List[QPersistentModelIndex] = []
        self.fetchingMore = False
        assert(self.model)

        self.model.columnsAboutToBeInserted.connect(self.runAllTests)
        self.model.columnsAboutToBeRemoved.connect(self.runAllTests)
        self.model.columnsInserted.connect(self.runAllTests)
        self.model.columnsRemoved.connect(self.runAllTests)
        self.model.dataChanged.connect(self.runAllTests)
        self.model.headerDataChanged.connect(self.runAllTests)
        self.model.layoutAboutToBeChanged.connect(self.runAllTests)
        self.model.layoutChanged.connect(self.runAllTests)
        self.model.modelReset.connect(self.runAllTests)
        self.model.rowsAboutToBeInserted.connect(self.runAllTests)
        self.model.rowsAboutToBeRemoved.connect(self.runAllTests)
        self.model.rowsInserted.connect(self.runAllTests)
        self.model.rowsRemoved.connect(self.runAllTests)

        # Special checks for inserting/removing
        self.model.layoutAboutToBeChanged.connect(self.layoutAboutToBeChanged)
        self.model.layoutChanged.connect(self.layoutChanged)
        self.model.rowsAboutToBeInserted.connect(self.rowsAboutToBeInserted)
        self.model.rowsAboutToBeRemoved.connect(self.rowsAboutToBeRemoved)
        self.model.rowsInserted.connect(self.rowsInserted)
        self.model.rowsRemoved.connect(self.rowsRemoved)
        self.runAllTests()

    def nonDestructiveBasicTest(self) ->None:
        """
        nonDestructiveBasicTest tries to call a number of the basic functions (not all)
        to make sure the model doesn't outright segfault, testing the functions that makes sense.
        """
        assert(self.model.buddy(QModelIndex()) == QModelIndex())
        self.model.canFetchMore(QModelIndex())
        assert(self.model.columnCount(QModelIndex()) >= 0)
        assert(self.model.data(QModelIndex(), Qt.ItemDataRole.DisplayRole) is None)
        self.fetchingMore = True
        self.model.fetchMore(QModelIndex())
        self.fetchingMore = False
        flags = self.model.flags(QModelIndex())
        assert(flags & Qt.ItemFlag.ItemIsEnabled == Qt.ItemFlag.ItemIsEnabled or
               flags & Qt.ItemFlag.ItemIsEnabled == 0)
        self.model.hasChildren(QModelIndex())
        self.model.hasIndex(0, 0)
        self.model.headerData(0, Qt.Orientation.Horizontal, Qt.ItemDataRole.DisplayRole)
        self.model.index(0, 0, QModelIndex())
        self.model.itemData(QModelIndex())
        cache = None
        self.model.match(QModelIndex(), -1, cache)
        self.model.mimeTypes()
        assert(self.model.parent(QModelIndex()) == QModelIndex())
        assert(self.model.rowCount(QModelIndex()) >= 0)
        variant = None
        self.model.setData(QModelIndex(), variant, -1)
        self.model.setHeaderData(-1, Qt.Orientation.Horizontal, None)
        self.model.setHeaderData(0, Qt.Orientation.Horizontal, None)
        self.model.setHeaderData(999999, Qt.Orientation.Horizontal, None)
        self.model.sibling(0, 0, QModelIndex())
        self.model.span(QModelIndex())
        self.model.supportedDropActions()

    def rowCount(self) ->None:
        """
        Tests self.model's implementation of QAbstractItemModel::rowCount() and hasChildren()

        self.models that are dynamically populated are not as fully tested here.
        """
        # check top row
        topindex = self.model.index(0, 0, QModelIndex())
        rows = self.model.rowCount(topindex)
        assert(rows >= 0)
        if rows > 0:
            assert(self.model.hasChildren(topindex))

        secondlvl = self.model.index(0, 0, topindex)
        if secondlvl.isValid():
            # check a row count where parent is valid
            rows = self.model.rowCount(secondlvl)
            assert(rows >= 0)
            if rows > 0:
                assert(self.model.hasChildren(secondlvl))

        # The self.models rowCount() is tested more extensively in checkChildren,
        # but this catches the big mistakes

    def columnCount(self) ->None:
        """
        Tests self.model's implementation of QAbstractItemModel::columnCount() and hasChildren()
        """
        # check top row
        topidx = self.model.index(0, 0, QModelIndex())
        assert(self.model.columnCount(topidx) >= 0)

        # check a column count where parent is valid
        childidx = self.model.index(0, 0, topidx)
        if childidx.isValid():
            assert(self.model.columnCount(childidx) >= 0)

        # columnCount() is tested more extensively in checkChildren,
        # but this catches the big mistakes

    def hasIndex(self) ->None:
        """
        Tests self.model's implementation of QAbstractItemModel::hasIndex()
        """
        # Make sure that invalid values returns an invalid index
        assert(self.model.hasIndex(-2, -2) == False)
        assert(self.model.hasIndex(-2, 0) == False)
        assert(self.model.hasIndex(0, -2) == False)

        rows = self.model.rowCount(QModelIndex())
        cols = self.model.columnCount(QModelIndex())

        # check out of bounds
        assert(self.model.hasIndex(rows, cols) == False)
        assert(self.model.hasIndex(rows + 1, cols + 1) == False)

        if rows > 0:
            assert(self.model.hasIndex(0, 0))

        # hasIndex() is tested more extensively in checkChildren()
        # but this catches the big mistakes

    def index(self) ->None:
        """
        Tests self.model's implementation of QAbstractItemModel::index()
        """
        # Make sure that invalid values returns an invalid index
        assert(self.model.index(-2, -2, QModelIndex()) == QModelIndex())
        assert(self.model.index(-2, 0, QModelIndex()) == QModelIndex())
        assert(self.model.index(0, -2, QModelIndex()) == QModelIndex())

        rows = self.model.rowCount(QModelIndex())
        cols = self.model.columnCount(QModelIndex())

        if rows == 0:
            return

        # Catch off by one errors
        assert(self.model.index(rows, cols, QModelIndex()) == QModelIndex())
        assert(self.model.index(0, 0, QModelIndex()).isValid())

        # Make sure that the same index is *always* returned
        idx1 = self.model.index(0, 0, QModelIndex())
        idx2 = self.model.index(0, 0, QModelIndex())
        assert(idx1 == idx2)

        # index() is tested more extensively in checkChildren()
        # but this catches the big mistakes

    def test_parent(self) ->None:
        """
        Tests self.model's implementation of QAbstractItemModel::parent()
        """
        # Make sure the self.model wont crash and will return an invalid QModelIndex
        # when asked for the parent of an invalid index
        assert(self.model.parent(QModelIndex()) == QModelIndex())

        if self.model.rowCount(QModelIndex()) == 0:
            return

        # Column 0              | Column 1  |
        # QtCore.Qself.modelIndex()         |           |
        #    \- topidx          | topidx1   |
        #         \- childix    | childidx1 |

        # Common error test #1, make sure that a top level index has a parent
        # that is an invalid QtCore.Qself.modelIndex
        topidx = self.model.index(0, 0, QModelIndex())
        assert(self.model.parent(topidx) == QModelIndex())

        # Common error test #2, make sure that a second level index has a parent
        # that is the first level index
        if self.model.rowCount(topidx) > 0:
            childidx = self.model.index(0, 0, topidx)
            assert(self.model.parent(childidx) == topidx)

        # Common error test #3, the second column should NOT have the same children
        # as the first column in a row
        # Usually the second column shouldn't have children
        topidx1 = self.model.index(0, 1, QModelIndex())
        if self.model.rowCount(topidx1) > 0:
            childidx = self.model.index(0, 0, topidx)
            childidx1 = self.model.index(0, 0, topidx1)
            assert(childidx != childidx1)

        # Full test, walk n levels deep through the self.model making sure that all
        # parent's children correctly specify their parent
        self.checkChildren(QModelIndex())

    def testRoleDataType(self, role:Qt.ItemDataRole, expectedType:Type[Any]) ->None:
        """Tests implementation if model.data() for role"""
        model = self.model
        result = model.data(model.index(0, 0, QModelIndex()), role)
        if result is not None and not isinstance(result, expectedType):
            raise Exception(f'returned data type is {type(result)}, should be {expectedType}')

    def testRoleDataValues(self, role:Qt.ItemDataRole, asserter:Any) ->None:
        """Tests implementation if model.data() for role.
        asserter is a function returning True or False"""
        model = self.model
        result = model.data(model.index(0, 0, QModelIndex()), role)
        if result is not None and not asserter(result):
            raise Exception('returned value {} is wrong')

    def data(self) ->None:
        """
        Tests self.model's implementation of QAbstractItemModel::data()
        """
        # Invalid index should return an invalid qvariant
        assert not isValid(self.model.data(QModelIndex(), Qt.ItemDataRole.DisplayRole))

        if self.model.rowCount(QModelIndex()) == 0:
            return

        # A valid index should have a valid data
        assert isValid(self.model.index(0, 0, QModelIndex()))

        # shouldn't be able to set data on an invalid index
        assert(
            self.model.setData(QModelIndex(),
                               "foo",
                               Qt.ItemDataRole.DisplayRole) == False)

        # General Purpose roles that should return a QString
        self.testRoleDataType(Qt.ItemDataRole.ToolTipRole, str)
        self.testRoleDataType(Qt.ItemDataRole.StatusTipRole, str)
        self.testRoleDataType(Qt.ItemDataRole.WhatsThisRole, str)
        self.testRoleDataType(Qt.ItemDataRole.SizeHintRole, QSize)
        self.testRoleDataType(Qt.ItemDataRole.FontRole, QFont)
        self.testRoleDataType(Qt.ItemDataRole.ForegroundRole, QColor)
        self.testRoleDataType(Qt.ItemDataRole.BackgroundRole, QColor)

        # Check that the alignment is one we know about
        self.testRoleDataValues(
            Qt.ItemDataRole.TextAlignmentRole,
            lambda x: x == (x & int(Qt.AlignmentFlag.AlignHorizontal_Mask | Qt.AlignmentFlag.AlignVertical_Mask)))

        # General Purpose roles that should return a QColor
        # Check that the "check state" is one we know about.
        self.testRoleDataValues(
            Qt.ItemDataRole.CheckStateRole,
            lambda x: x in (Qt.CheckState.Unchecked, Qt.CheckState.PartiallyChecked, Qt.CheckState.Checked))

    def runAllTests(self) ->None:
        """run all tests after the model changed"""
        if not isAlive(self):
            return
        if self.fetchingMore:
            return
        self.nonDestructiveBasicTest()
        self.rowCount()
        self.columnCount()
        self.hasIndex()
        self.index()
        self.test_parent()
        self.data()

    def rowsAboutToBeInserted(self, parent:QModelIndex, start:int, unusedEnd:int) ->None:
        """
        Store what is about to be inserted to make sure it actually happens
        """
        item:Dict[str, Any] = {}
        item['parent'] = parent
        item['oldSize'] = self.model.rowCount(parent)
        item['last'] = self.model.data(self.model.index(start - 1, 0, parent))
        item['next'] = self.model.data(self.model.index(start, 0, parent))
        self.insert.append(item)

    def rowsInserted(self, parent:QModelIndex, start:int, end:int) ->None:
        """
        Confirm that what was said was going to happen actually did
        """
        item = self.insert.pop()
        assert(item['parent'] == parent)
        assert(item['oldSize'] + (end - start + 1)
               == self.model.rowCount(parent))
        assert(item['last'] == self.model.data(
            self.model.index(start - 1, 0, item['parent'])))

        # if item['next'] != self.model.data(self.model.index(end+1, 0, item['parent'])):
        #   qDebug << start << end
        #   for i in range(0, self.model.rowCount(QModelIndex())):
        #       qDebug << self.model.index(i, 0).data().toString()
        # qDebug() << item['next'] << self.model.data(model.index(end+1, 0,
        # item['parent']))

        assert(item['next'] == self.model.data(
            self.model.index(end + 1, 0, item['parent'])))

    def rowsAboutToBeRemoved(self, parent:QModelIndex, start:int, end:int) ->None:
        """
        Store what is about to be inserted to make sure it actually happens
        """
        item:Dict[str, Any] = {}
        item['parent'] = parent
        item['oldSize'] = self.model.rowCount(parent)
        item['last'] = self.model.data(self.model.index(start - 1, 0, parent))
        item['next'] = self.model.data(self.model.index(end + 1, 0, parent))
        self.remove.append(item)

    def rowsRemoved(self, parent:QModelIndex, start:int, end:int) ->None:
        """
        Confirm that what was said was going to happen actually did
        """
        item = self.remove.pop()
        assert(item['parent'] == parent)
        assert(item['oldSize'] - (end - start + 1)
               == self.model.rowCount(parent))
        assert(item['last'] == self.model.data(
            self.model.index(start - 1, 0, item['parent'])))
        assert(item['next'] == self.model.data(
            self.model.index(start, 0, item['parent'])))

    def layoutAboutToBeChanged(self) ->None:
        """
        Store what is about to be changed
        """
        for i in range(0, max(0, min(self.model.rowCount(), 100))):
            self.changing.append(QPersistentModelIndex(self.model.index(i, 0)))

    def layoutChanged(self) ->None:
        """
        Confirm that what was said was going to happen actually did
        """
        for change in self.changing:
            assert(
                change == self.model.index(change.row(),
                                           change.column(),
                                           change.parent()))
        self.changing = []

    def checkChildren(self, parent:QModelIndex, depth:int=0) ->None:
        """
        Called from parent() test.

        A self.model that returns an index of parent X should also return X when asking
        for the parent of the index

        This recursive function does pretty extensive testing on the whole self.model in an
        effort to catch edge cases.

        This function assumes that rowCount(QModelIndex()), columnCount(QModelIndex()) and index() already work.
        If they have a bug it will point it out, but the above tests should have already
        found the basic bugs because it is easier to figure out the problem in
        those tests then this one
        """
        # First just try walking back up the tree.
        parentIdx = parent
        while parentIdx.isValid():
            parentIdx = parentIdx.parent()

        # For self.models that are dynamically populated
        if self.model.canFetchMore(parent):
            self.fetchingMore = True
            self.model.fetchMore(parent)
            self.fetchingMore = False

        rows = self.model.rowCount(parent)
        cols = self.model.columnCount(parent)

        if rows > 0:
            assert(self.model.hasChildren(parent))

        # Some further testing against rows(), columns, and hasChildren()
        assert(rows >= 0)
        assert(cols >= 0)

        if rows > 0:
            assert(self.model.hasChildren(parent))

        # qDebug() << "parent:" << self.model.data(parent).toString() << "rows:" << rows
        #          << "columns:" << cols << "parent column:" << parent.column()

        assert(self.model.hasIndex(rows + 1, 0, parent) == False)
        for row in range(0, rows):
            if self.model.canFetchMore(parent):
                self.fetchingMore = True
                self.model.fetchMore(parent)
                self.fetchingMore = False
            assert(self.model.hasIndex(row, cols + 1, parent) == False)
            for column in range(0, cols):
                assert(self.model.hasIndex(row, column, parent))
                index = self.model.index(row, column, parent)
                # rowCount(QModelIndex()) and columnCount(QModelIndex()) said
                # that it existed...
                assert(index.isValid())

                # index() should always return the same index when called twice
                # in a row
                modIdx = self.model.index(row, column, parent)
                assert(index == modIdx)

                # Make sure we get the same index if we request it twice in a
                # row
                idx1 = self.model.index(row, column, parent)
                idx2 = self.model.index(row, column, parent)
                assert(idx1 == idx2)

                # Some basic checking on the index that is returned
                # assert( index.model() == self.model )
                # This raises an error that is not part of the qbzr code.
                # see
                # https://www.riverbankcomputing.com/pipermail/pyqt/2011-February/029300.html
                assert(index.row() == row)
                assert(index.column() == column)
                # While you can technically return a QVariant usually this is a sign
                # if an bug in data() Disable if this really is ok in your
                # self.model
                assert isValid(self.model.data(index, Qt.ItemDataRole.DisplayRole))

                # if the next test fails here is some somewhat useful debug you play with
                # if self.model.parent(index) != parent:
                #   qDebug() << row << column << depth << self.model.data(index).toString()
                #        << self.model.data(parent).toString()
                #   qDebug() << index << parent << self.model.parent(index)
                # And a view that you can even use to show the self.model
                # view = QtGui.QTreeView()
                # view.setself.model(model)
                # view.show()
                #

                # Check that we can get back our real parent
                parentIdx = self.model.parent(index)
                assert(parentIdx.internalId() == parent.internalId())
                assert(parentIdx.row() == parent.row())

                # recursively go down the children
                if self.model.hasChildren(index) and depth < 10:
                    # qDebug() << row << column << "hasChildren" <<
                    # self.model.rowCount(index)
                    self.checkChildren(index, depth + 1)
                # else:
                #   if depth >= 10:
                #       qDebug() << "checked 10 deep"

                # Make sure that after testing the children that the index
                # doesn't change
                newIdx = self.model.index(row, column, parent)
                assert(index == newIdx)



# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import List, Optional, Union, TYPE_CHECKING, cast

from qt import Qt, QAbstractTableModel, QModelIndex, QPersistentModelIndex, QSize
from qt import QWidget, QLineEdit, QVBoxLayout, QColor, QAbstractItemView

from log import i18n, logDebug
from guiutil import MJTableView, decorateWindow
from statesaver import StateSaver
from message import ChatMessage
from common import Debug, Internal
from modeltest import ModelTest

if TYPE_CHECKING:
    from twisted.python.failure import Failure
    from qt import QObject
    from client import ClientTable
    from scene import GameScene
    from humanclient import HumanClient

class ChatModel(QAbstractTableModel):

    """a model for the chat view"""

    def __init__(self, parent:Optional['QObject']=None) ->None:
        super().__init__(parent)
        self.chatLines:List[ChatMessage] = []

    def headerData(self, section:int, orientation:Qt.Orientation,
        role:int=Qt.ItemDataRole.DisplayRole) ->Union[int, str, None]:
        """show header"""
        if role == Qt.ItemDataRole.TextAlignmentRole:
            if orientation == Qt.Orientation.Horizontal:
                if section == 1:
                    return int(Qt.AlignmentFlag.AlignRight)
                return int(Qt.AlignmentFlag.AlignLeft)
        if orientation != Qt.Orientation.Horizontal:
            return int(section + 1)
        if role != Qt.ItemDataRole.DisplayRole:
            return None
        result = ''
        if section < self.columnCount():
            result = [i18n('Time'), i18n('Player'), i18n('Message')][section]
        return result

    def rowCount(self, parent:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->int:
        """how many lines are in the model?"""
        if parent.isValid():
            # similar in tables.py
            # we have only top level items
            return 0
        return len(self.chatLines)

    def columnCount(self, unusedParent:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->int:
        """for now we only have time, who, message"""
        return 3

    def data(self, index:Union[QModelIndex,QPersistentModelIndex],
        role:int=Qt.ItemDataRole.DisplayRole) ->Union[int, str, QColor, None]:
        """score table"""
        result:Union[int, str, QColor, None] = None
        if role == Qt.ItemDataRole.TextAlignmentRole:
            if index.column() == 1:
                return int(Qt.AlignmentFlag.AlignRight)
            return int(Qt.AlignmentFlag.AlignLeft)
        if index.isValid() and (0 <= index.row() < len(self.chatLines)):
            chatLine = self.chatLines[index.row()]
            if role == Qt.ItemDataRole.DisplayRole and index.column() == 0:
                local = chatLine.localtimestamp()
                result = f'{int(local.hour):02}:{int(local.minute):02}:{int(local.second):02}'
            elif role == Qt.ItemDataRole.DisplayRole and index.column() == 1:
                result = chatLine.fromUser
            elif role == Qt.ItemDataRole.DisplayRole and index.column() == 2:
                result = i18n(chatLine.message)
            elif role == Qt.ItemDataRole.ForegroundRole and index.column() == 2:
                palette = Internal.app.palette()
                color = 'blue' if chatLine.isStatusMessage else palette.windowText(
                )
                result = QColor(color)
        return result

    def appendLine(self, line:ChatMessage) ->None:
        """insert a chatline"""
        old_rowCount = self.rowCount()
        self.beginInsertRows(QModelIndex(), old_rowCount, old_rowCount)
        self.chatLines.append(line)
        self.endInsertRows()


class ChatView(MJTableView):

    """define a minimum size"""

    def __init__(self) ->None:
        super().__init__()

    def sizeHint(self) ->QSize:
        """sizeHint"""
        return QSize(400, 100)

    def minimumSizeHint(self) ->QSize:
        """minimumSizeHint"""
        return self.sizeHint()


class ChatWindow(QWidget):

    """a widget for showing chat messages"""

    def __init__(self, table:'ClientTable') ->None:
        super().__init__(None)
        assert table.client.connection
        self.table = table
        table.chatWindow = self
        self.setObjectName('chatWindow')
        title = i18n(
            'Chat on table %1 at %2',
            table.tableid,
            table.client.connection.url)
        decorateWindow(self, title)
        self.messageView = ChatView()
        self.messageView.setModel(ChatModel())
        self.messageView.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.messageView.setShowGrid(False)
        self.messageView.setWordWrap(False)
        self.messageView.setSelectionMode(QAbstractItemView.SelectionMode.NoSelection)
        if Debug.modelTest:
            if model := self.messageView.model():
                self.debugModelTest = ModelTest(model, self.messageView)
        self.edit = QLineEdit()
        layout = QVBoxLayout()
        layout.addWidget(self.messageView)
        layout.addWidget(self.edit)
        self.setLayout(layout)
        self.edit.returnPressed.connect(self.sendLine)
        self.edit.setFocus()
        self.show()
        StateSaver(self)

    def show(self) ->None:
        """not only show but also restore and raise"""
        self.activateWindow()
        self.setWindowState(cast(Qt.WindowState, self.windowState() & ~Qt.WindowState.WindowMinimized))
        self.raise_()
        super().show()

    def isVisible(self) ->bool:
        """not only visible but also not minimized"""
        return QWidget.isVisible(self) and not self.windowState() & Qt.WindowState.WindowMinimized

    def kill(self) ->None:
        """hide and null on table"""
        if Debug.chat:
            logDebug(f'chat.kill for {self} on table {self.table}')
        self.hide()
        assert self.table
        self.table.chatWindow = None

    def sendLine(self, line:Optional[str]=None, isStatusMessage:bool=False) ->None:
        """send line to others. Either the edited line or parameter line."""
        if line is None:
            line = self.edit.text()
            self.edit.clear()
        if line:
            assert self.table
            assert self.table.client
            if Debug.chat:
                logDebug(f'sending line {line} to others')
            msg = ChatMessage(
                self.table.tableid,
                self.table.client.name,
                line,
                isStatusMessage)
            cast('HumanClient', self.table.client).sendChat(msg).addErrback(self.chatError)

    def chatError(self, result:'Failure') ->None:
        """tableList may already have gone away"""
        assert self.table
        assert self.table.client
        cast('HumanClient', self.table.client).tableError(result)

    def leave(self) ->None:
        """leaving the chat"""
        self.hide()

    def receiveLine(self, chatLine:ChatMessage) ->None:
        """show a new line in protocol"""
        self.show()
        if model := self.messageView.model():
            cast(ChatModel, model).appendLine(chatLine)
            for row in range(model.rowCount()):
                self.messageView.setRowHeight(
                    row,
                    self.messageView.fontMetrics().height())
        self.messageView.resizeColumnsToContents()
        self.messageView.scrollToBottom()



"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

partially based on C++ code from:
Copyright (C) 2006 Mauricio Piacentini <mauricio@tabuleiro.com>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import cast

from qt import QWidget, QHBoxLayout, QLineEdit
from qt import QSize, QVBoxLayout, QSizePolicy, QGroupBox, QListWidget
from qt import QFormLayout, QLabel, QSpacerItem, QMetaObject
from tile import Tile
from tileset import Tileset
from uitile import UITile
from board import Board, FittingView
from scene import SceneWithFocusRect
from common import Internal
from wind import Wind
from animation import AnimationSpeed
from mi18n import i18nc


class TilesetSelector(QWidget):

    # pylint:disable=too-many-instance-attributes

    """presents all available tiles with previews"""

    def __init__(self, parent:QWidget) ->None:
        super().__init__(parent)

        self.tilesetNameList:'QListWidget'
        self.tilesetAuthor:'QLabel'
        self.tilesetContact:'QLabel'
        self.tilesetDescription:'QLabel'
        self.tilesetPreview:'QWidget'

        assert Internal.Preferences
        self.setupUi()
        self.kcfg_tilesetName = QLineEdit(self)
        self.kcfg_tilesetName.setVisible(False)
        self.kcfg_tilesetName.setObjectName('kcfg_tilesetName')

        self.tileScene = SceneWithFocusRect()
        self.tileView = FittingView()
        self.tileView.setScene(self.tileScene)
        _ = Internal.Preferences.tilesetName
        assert isinstance(_, str)
        self.tileset = Tileset(_)
        self.uiTiles = [UITile(Tile('w' + s.char.lower())) for s in Wind.all4]
        self.board = Board(2, 2, self.tileset)
        self.tileScene.addItem(self.board)
        self.tileView.setParent(self.tilesetPreview)
        layout = QHBoxLayout(self.tilesetPreview)
        layout.addWidget(self.tileView)
        for idx, offsets in enumerate([(0, 0), (0, 1), (1, 0), (1, 1)]):
            self.uiTiles[idx].setBoard(
                self.board,
                *offsets)
            self.uiTiles[idx].focusable = False
        self.setUp()

    def setupUi(self) ->None:
        """created by pyuic from old tilesetselector.ui and adapted for direct use"""

        # pylint:disable=too-many-statements
        self.setObjectName("TilesetSelector")
        self.resize(497, 446)
        self.setMaximumSize(QSize(800, 600))
        self.vboxlayout = QVBoxLayout(self)
        self.vboxlayout.setContentsMargins(0, 0, 0, 0)
        self.vboxlayout.setObjectName("vboxlayout")
        self.hboxlayout = QHBoxLayout()
        self.hboxlayout.setContentsMargins(0, 0, 0, 0)
        self.hboxlayout.setObjectName("hboxlayout")
        self.tilesetNameList = QListWidget(self)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(2)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tilesetNameList.sizePolicy().hasHeightForWidth())
        self.tilesetNameList.setSizePolicy(sizePolicy)
        self.tilesetNameList.setMinimumSize(QSize(120, 0))
        self.tilesetNameList.setObjectName("tilesetNameList")
        self.hboxlayout.addWidget(self.tilesetNameList)
        self.groupBox_2 = QGroupBox(self)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_2.sizePolicy().hasHeightForWidth())
        self.groupBox_2.setSizePolicy(sizePolicy)
        self.groupBox_2.setObjectName("groupBox_2")
        self.vboxlayout1 = QVBoxLayout(self.groupBox_2)
        self.vboxlayout1.setContentsMargins(0, 0, 0, 0)
        self.vboxlayout1.setObjectName("vboxlayout1")
        self.tilesetPreview = QWidget(self.groupBox_2)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tilesetPreview.sizePolicy().hasHeightForWidth())
        self.tilesetPreview.setSizePolicy(sizePolicy)
        self.tilesetPreview.setMinimumSize(QSize(120, 160))
        self.tilesetPreview.setObjectName("tilesetPreview")
        self.vboxlayout1.addWidget(self.tilesetPreview)
        self.hboxlayout.addWidget(self.groupBox_2)
        self.vboxlayout.addLayout(self.hboxlayout)
        self.groupBox = QGroupBox(self)
        self.groupBox.setObjectName("groupBox")
        self.formLayout = QFormLayout(self.groupBox)
        self.formLayout.setObjectName("formLayout")
        self.labelAuthor = QLabel(self.groupBox)
        self.labelAuthor.setObjectName("labelAuthor")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.labelAuthor)
        self.tilesetAuthor = QLabel(self.groupBox)
        self.tilesetAuthor.setObjectName("tilesetAuthor")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.tilesetAuthor)
        self.labelContact = QLabel(self.groupBox)
        self.labelContact.setObjectName("labelContact")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.labelContact)
        self.tilesetContact = QLabel(self.groupBox)
        self.tilesetContact.setObjectName("tilesetContact")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.tilesetContact)
        self.labelDescription = QLabel(self.groupBox)
        self.labelDescription.setObjectName("labelDescription")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.LabelRole, self.labelDescription)
        self.tilesetDescription = QLabel(self.groupBox)
        self.tilesetDescription.setObjectName("tilesetDescription")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.tilesetDescription)
        self.vboxlayout.addWidget(self.groupBox)
        spacerItem = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.vboxlayout.addItem(spacerItem)

        self.retranslateUi()
        QMetaObject.connectSlotsByName(self)

    def retranslateUi(self) ->None:
        """created by pyuic from old tilesetselector.ui and adapted for direct use"""
        self.groupBox_2.setTitle(i18nc("TilesetSelector", "Preview"))
        self.groupBox.setTitle(i18nc("TilesetSelector", "Properties"))
        self.labelAuthor.setText(i18nc("TilesetSelector", "Author:"))
        self.labelContact.setText(i18nc("TilesetSelector", "Contact:"))
        self.labelDescription.setText(i18nc("TilesetSelector", "Description:"))

    def setUp(self) ->None:
        """set-up the selector"""

        assert Internal.Preferences
        # The lineEdit widget holds our tileset path, but the user does
        # not manipulate it directly
        self.kcfg_tilesetName.hide()

        self.tilesetNameList.currentRowChanged.connect(self.tilesetRowChanged)
        self.kcfg_tilesetName.textChanged.connect(self.tilesetNameChanged)

        Tileset.loadAll()
        # list default tileset first
        self.tilesetList = Tileset.available()
        for aset in self.tilesetList:
            self.tilesetNameList.addItem(aset.name)
        self.kcfg_tilesetName.setText(cast(str, Internal.Preferences.tilesetName))

    def tilesetNameChanged(self, name:str) ->None:
        """the name changed: update the current row"""
        igrindex = 0
        for idx, aset in enumerate(self.tilesetList):
            if aset.desktopFileName == name:
                igrindex = idx
                break
        self.tilesetNameList.setCurrentRow(igrindex)

    def tilesetRowChanged(self) ->None:
        """user selected a new tileset, update our information about it and
        paint preview"""
        selTileset = cast(Tileset, self.tilesetList[self.tilesetNameList.currentRow()])
        self.kcfg_tilesetName.setText(selTileset.desktopFileName)
        self.tilesetAuthor.setText(selTileset.author)
        self.tilesetContact.setText(selTileset.authorEmail)
        self.tilesetDescription.setText(selTileset.description)
        with AnimationSpeed():
            self.board.tileset = selTileset
